!-----------------------------------------------------------------------------------!
!                                                                                   !
!                            QUASIATOMIC ORBITAL (QO)                               !
!                                                                                   !
!   Description:  Quasiatomic Wannier Functions in Molecules and Solids             !
!                                                                                   !
!   Author:       Xiaofeng Qian  (MIT, Cambridge, MA, USA)                          !
!                                                                                   !
!   Reference:   "Quasiatomic orbitals for ab initio tight-binding analysis",       !
!                 Xiaofeng Qian, Ju Li, Liang Qi, Cai-Zhuang Wang,                  !
!                 Tzu-Liang Chan, Yong-Xin Yao, Kai-Ming Ho, and Sidney Yip,        !
!                 Physical Review B 78, 245112 (2008)                               !
!                                                                                   !
!   Reference (in BIBTEX format):                                                   !
!                                                                                   !
!     @ARTICLE{QianLQWCYHY08,                                                       !
!          author    = {Qian, Xiaofeng and Li, Ju and Qi, Liang and Wang, C. Z. and !
!                       Chan, T. L. and Yao, Y. X. and Ho, K. M. and Yip, Sidney},  !
!          title     = {Quasiatomic orbitals for ab initio tight-binding analysis}, ! 
!          journal   = {Phys. Rev. B},                                              !
!          year	     = {2008},                                                      !
!          volume    = {78},                                                        !
!          pages     = {245112}                                                     !
!     }                                                                             !
!                                                                                   !
!   Please cite the above reference if you use the code or part of the code.        !  
!                                                                                   !
!   Copyright (C) 2009 Xiaofeng Qian                                                !
!                                                                                   !
!   This file is distributed under the terms of the                                 !
!   GNU General Public License. See the file `License'                              !
!   in the root directory of the present distribution,                              !
!   or http://www.gnu.org/copyleft/gpl.txt .                                        !
!                                                                                   !
!-----------------------------------------------------------------------------------!
!
#include "f_defs.h"
!
!-----------------------------------------------------------------------
subroutine QO_init_at_1_so()
  !-----------------------------------------------------------------------
  !
  ! This routine computes a table with the radial Fourier transform 
  ! of the atomic wavefunctions.
  !
  use QO_module,  only : tab_at_lm, nchix_max, QO_PseudoatomicOrbital_rescalingfactor
  use io_global,  only : stdout, ionode
  USE uspp_param, ONLY : upf 
  USE kinds,      ONLY : dp
  USE atom,       ONLY : rgrid, msh
  USE constants,  ONLY : fpi
  USE cell_base,  ONLY : omega
  USE ions_base,  ONLY : ntyp => nsp
  USE us,         ONLY : tab_at, nqx, dq
  USE mp_global,  ONLY : intra_pool_comm 
  USE mp,         ONLY : mp_sum
  !
  implicit none
  !
  integer :: n_starting_wfc, nt, nb, iq, ir, l, startq, lastq, ndm, ii, jj, m
  !
  real(DP), allocatable :: aux0(:), aux (:), chi0(:), chi(:), vchi (:), aux_scaling(:)
  real(DP), allocatable :: chi0_collect(:), chi_collect(:), vchi_collect(:), aux_scaling_collect(:)
  real(DP) :: vqint, pref, q
  real(dp) :: orb_integral_old, orb_integral_new, rc, smearing_factor
  logical  :: ldebug
  
  call start_clock ('QO_init_at_1')

  ndm = MAXVAL(msh(1:ntyp))
  
  allocate (aux0(ndm), aux(ndm), chi0(ndm), chi(ndm), vchi(ndm), aux_scaling(ndm))
  allocate (chi0_collect(nqx), chi_collect(nqx), vchi_collect(nqx), aux_scaling_collect(nqx))
  
  !
  ! chiq = radial fourier transform of atomic orbitals chi
  !
  pref = fpi/sqrt(omega)
  !
  ! needed to normalize atomic wfcs (not a bad idea in general and 
  ! necessary to compute correctly lda+U projections)
  !
  call divide (nqx, startq, lastq)  
!!$  startq = 1
!!$  lastq  = nqx
  
  tab_at_lm(:,:,:) = 0.d0
  m = 0

  do nt = 1, ntyp
     
     ii = 1
     
     do nb = 1, upf(nt)%nwfc
        if (upf(nt)%oc (nb) >= 0.d0) then
           l = upf(nt)%lchi (nb)
           
           if (ionode) write(stdout, '(A, 4I)') '[init_at_1] nt, nb, l, ii = ', nt, nb, l, ii
           
           aux0 = 0.d0
           chi0 = 0.d0
           
           ldebug = .true.
           ! calculate <phi(r)|S|phi(r)>
           do jj = 1, msh(nt)
              aux0(jj) = upf(nt)%chi(jj, nb) * upf(nt)%chi(jj, nb)
              chi0(jj) = upf(nt)%chi(jj, nb)              
           end do
           call simpson (msh(nt), aux0, rgrid(nt)%rab, orb_integral_old)
           if (ionode) write(stdout, *) ' <old>: ', nt, nb, l, 0, orb_integral_old
           
!           do m = 1, 2*l+1
              
              smearing_factor = 1.d0
              rc = 10.d0 ! bohr
              rc = QO_PseudoatomicOrbital_rescalingfactor(nt, ii)
              aux_scaling = 0.d0
              do jj= 1, msh(nt)
                 !
                 ! option 1: re-scaled by exponential function
                 aux_scaling(jj) = exp( - QO_PseudoatomicOrbital_rescalingfactor(nt, ii) * rgrid(nt)%r(jj) )
                 !
                 ! option 2: re-scaled by Fermi-Dirac function
                 !! aux_scaling(jj) = 1.d0 / (exp( (rgrid(nt)%r(jj) - rc)/smearing_factor ) + 1.d0)
                 !
              end do
              
              do jj = 1, msh(nt)
                 aux(jj) = aux0(jj) * aux_scaling(jj) * aux_scaling(jj)                      
              end do
              call simpson (msh(nt), aux, rgrid(nt)%rab, orb_integral_new)
              
              if (ionode) write(stdout, *) ' <new>: ', nt, nb, l, m, orb_integral_new
              if (ionode) write(stdout, *) ' nt, ii, rescale = ', nt, ii, QO_PseudoatomicOrbital_rescalingfactor(nt, ii)
              
              do jj = 1, msh(nt)
                 chi(jj) = chi0(jj) * &
                      aux_scaling(jj) * &
                      dsqrt( orb_integral_old / orb_integral_new)
              end do
              
              if (ldebug) then
                 write(stdout, *) '------------------------------------------------------'
                 write(stdout, *) nt, nb, l, m
                 write(stdout, *) 'i   r(i)   s(i)  chi0(i)   chi(i)'
                 do jj = 1, msh(nt)
                    write(stdout, '(I10, 4F16.10)') jj, rgrid(nt)%r(jj), aux_scaling(jj), chi0(jj), chi(jj)
                 end do
                 write(stdout, *) '------------------------------------------------------'
                 ldebug = .false.
              end if
              
!!$           do jj = 1, msh(nt)
!!$              aux(jj) = upf(nt)%chi(jj, nb) * upf(nt)%chi(jj, nb)
!!$           end do
!!$           call simpson (msh(nt), aux, rgrid(nt)%rab, orb_integral_new)           
              
              aux = 0.d0
              do iq = startq, lastq
                 
                 q = dq * (iq - 1)
                 call sph_bes (msh(nt), rgrid(nt)%r, q, l, aux)
                 do ir = 1, msh(nt)
                    !
                    ! old version without rescaling of Pseudo-AO
                    ! vchi(ir) = chi(ir,nb,nt) * aux(ir) * rgrid(nt)%r(ir)
                    !
                    ! new version for the case of QO
                    vchi(ir) = chi(ir) * aux(ir) * rgrid(nt)%r(ir) 
                    
                 enddo
                 call simpson (msh(nt), vchi, rgrid(nt)%rab, vqint)
                 tab_at_lm (iq, ii, nt) = vqint * pref
              enddo
              
              if (m== 2*l+1) then
                 upf(nt)%chi(1:msh(nt), nb) = chi(1:msh(nt))
              end if
              ii = ii + 1              
!           end do   ! do m = 1, 2*l+1
           
        endif   ! if ( oc(nb) >= 0.d0 )
     enddo   ! do nb = 1, upf(nt)%nwfc
  enddo
#ifdef __PARA
  ! recover over G parallelization (intra_pool)
  call mp_sum (tab_at_lm, intra_pool_comm)
#endif
  
  deallocate( aux0, aux, chi0, chi, vchi, aux_scaling )
  deallocate( chi0_collect, chi_collect, vchi_collect, aux_scaling_collect )
  
  call stop_clock ('QO_init_at_1')
  return
  
end subroutine QO_init_at_1

