!-----------------------------------------------------------------------------------!
!                                                                                   !
!                            QUASIATOMIC ORBITAL (QO)                               !
!                                                                                   !
!   Description:  Quasiatomic Wannier Functions in Molecules and Solids             !
!                                                                                   !
!   Author:       Xiaofeng Qian  (MIT, Cambridge, MA, USA)                          !
!                                                                                   !
!   Reference:   "Quasiatomic orbitals for ab initio tight-binding analysis",       !
!                 Xiaofeng Qian, Ju Li, Liang Qi, Cai-Zhuang Wang,                  !
!                 Tzu-Liang Chan, Yong-Xin Yao, Kai-Ming Ho, and Sidney Yip,        !
!                 Physical Review B 78, 245112 (2008)                               !
!                                                                                   !
!   Reference (in BIBTEX format):                                                   !
!                                                                                   !
!     @ARTICLE{QianLQWCYHY08,                                                       !
!          author    = {Qian, Xiaofeng and Li, Ju and Qi, Liang and Wang, C. Z. and !
!                       Chan, T. L. and Yao, Y. X. and Ho, K. M. and Yip, Sidney},  !
!          title     = {Quasiatomic orbitals for ab initio tight-binding analysis}, ! 
!          journal   = {Phys. Rev. B},                                              !
!          year	     = {2008},                                                      !
!          volume    = {78},                                                        !
!          pages     = {245112}                                                     !
!     }                                                                             !
!                                                                                   !
!   Please cite the above reference if you use the code or part of the code.        !  
!                                                                                   !
!   Copyright (C) 2009 Xiaofeng Qian                                                !
!                                                                                   !
!   This file is distributed under the terms of the                                 !
!   GNU General Public License. See the file `License'                              !
!   in the root directory of the present distribution,                              !
!   or http://www.gnu.org/copyleft/gpl.txt .                                        !
!                                                                                   !
!-----------------------------------------------------------------------------------!
!
module QO_module
  
  use kinds,      only : DP
  
  save
  
  type QO_input_type
     character(len=256) :: outdir
     character(len=256) :: prefix
     integer :: ao_lm
     real(dp):: ao_rescaling
     logical :: tightbinding_flag
     logical :: orbital_flag     
  end type QO_input_type
  
  ! =============================================================================================
  !
  ! input variables: begin ...
  !
  logical :: &
       QO_flag_construction    = .true.
  logical :: &                                                
       QO_flag_orthonormalized = .false.   ! .TRUE. if orthonormalized is performed
  logical :: &
       QO_flag_normalized      = .true.    ! .TRUE. if QO is normalized
  logical :: &
       QO_flag_H_diagonal      = .false.   ! .TRUE. if only using the diagonal element of Hamiltonian as eigenvalue
  logical :: &
       QO_flag_diagonalize_Hamiltonian = .true.  ! .TRUE. if we diagonalize \varepsilon matrix in the c_mk subspace
                                                 ! so that we can remove the matrix form of \varepsilon
  logical :: &
       QO_flag_output_orbital  = .false.   ! .TRUE. if generating XSF&NetCDF file for Quasiatomic Orbitals

  character(len=60) :: &
       QO_flag_output_orbital_format  = 'XSF'   ! 'xsf', 'netcdf', 'both'
  
  real(dp) :: QO_E_window_max = 0.d0         ! upper energy limit of bands to be kept frozen
  real(dp) :: QO_E_window_min = - 100000.d0  ! lower energy limit of bands to be kept frozen
  real(dp) :: QO_unboundstate_cutoff = 0.d0  ! cutoff for the unbound state which has small overlap with AO Bloch subspace
  integer  :: QO_transport_direction = 0     !  0: sum-over-alldirections (default)
                                             !  1: b1 (sum k-points on the directions of b2, b3)
                                             !  2: b2 (sum k-points on the directions of b1, b3)
                                             !  3: b3 (sum k-points on the directions of b1, b2)
                                             ! -1: do not sum k-points on any direction 
  
  real(dp) :: QO_flag_output_orbital_rcutoff = 0.0  ! Angstrom, for output netcdf file for a small region, default value = 0.0 for all region
  
  integer,  allocatable ::   &
       QO_PseudoatomicOrbital_nlm(:,:), &
       QO_PseudoatomicOrbital_nlm_effective(:), &
       QO_PseudoatomicOrbital_channel_add(:, :), &
       QO_PseudoatomicOrbital_channel_remove(:, :), &
       QO_PseudoatomicOrbital_number(:)  
  
  integer :: QO_plot_cell_dimension(3) = (/1, 1, 1/)  
  integer :: QO_plot_natom
  integer :: QO_plot_number_of_kpoints = 0
  integer,  allocatable  ::   QO_plot_atomindex(:)
  integer,  allocatable  ::   QO_plot_atomflag(:)
  real(dp), allocatable  ::   QO_plot_kpoints(:,:) ! not used
  integer,  allocatable  ::   QO_plot_kpoints_index(:)
  !
  ! input variables: end.
  !
  ! =============================================================================================
  
  
  
  
  ! =============================================================================================
  ! 
  ! internal variables: begin ...
  !
  integer  :: ncell, nkxyz, ncellx, ncelly, ncellz, ncellhalfx, ncellhalfy, ncellhalfz
  integer  :: ncell_plot = 1
  integer  :: ncellx_plot = 1
  integer  :: ncelly_plot = 1
  integer  :: ncellz_plot = 1
  integer  :: ncellhalfx_plot = 0
  integer  :: ncellhalfy_plot = 0
  integer  :: ncellhalfz_plot = 0
  integer  :: nchix_max = 30
  integer  :: nspin_general
  
  integer  :: nrxx_total, nrxxs_total
  
  integer  :: iunsatr, nwordatwfcr
  
  integer, allocatable  :: QO_nbndk(:)
  integer, allocatable  :: QO_nbndk_begin(:)
  integer, allocatable  :: QO_nbndk_end(:)
  integer, allocatable  :: QO_bandindex(:,:)
    
  complex(DP), allocatable :: &
       QO_wfc(:,:),           & ! |k+G><k+G|QO_Ii>
       QO_wfc_r(:,:,:),       & ! |R+r><R+r|QO_Ii>
       QO_Omega(:,:,:),       & ! <psi_nk | QO_Ii>
       QO_cmk(:,:),           & ! |k+G><k+G| c_mk>
       QO_Enmk(:,:,:),        & ! <c_mk| H |c_nk>
       QO_H(:,:,:,:),         & ! <QO_Ii(X_0) | H | QO_Jj(X_L)>
       QO_S(:,:,:,:),         & ! <QO_Ii(X_0) | S | QO_Jj(X_L)>
       QO_OEO(:,:,:),         & ! QO_Omega^{+} * QO_Enmk * QO_Omega (Ii, Jj, ik)
       QO_OO(:,:,:)             ! QO_Omega^{+} * QO_Omega (Ii, Jj, ik)
  
  real(DP), allocatable ::   &
       QO_charge(:,:,:,:),   &  ! Mulliken/Lowdin charge
       QO_PseudoatomicOrbital_rescalingfactor(:,:), &  ! rescacaling factor for AO projectors
       Bloch_projsum(:,:),   &  ! <psi_nk     | S | AO_Ii(k)  >
       AO_norm(:)               ! <AO_Ii|S|AO_Ii>

  real(DP), allocatable :: &
       Atom_position(:,:), &    ! in Angstrom: tau(1:dim3,1:natom) * alat * bohr_to_angstrom
       QO_position(:,:),&       ! in Angstrom: tau(1:dim3,1:natom) * alat * bohr_to_angstrom
       Xgrid(:,:), &
       Xgrid_total(:,:)
  
  real(DP), allocatable :: tab_at_lm(:,:,:)
  
  integer, allocatable :: &
       evc_unocc_index(:,:)
  
  real(DP), allocatable :: &
       evc_unocc_energy(:,:)
  
  real(DP), allocatable :: &
       vcut(:)
  
  integer, allocatable :: &
       QO_to_Atom(:), &   ! 
       Atom_to_QO(:, :), &   !        
  
  ! number of occupied bands at each k-point
  integer, allocatable :: nbnd_occ(:)
  
  logical :: &
       l_QO_initialized        = .false.   ! .TRUE. if QO module has already been initialized
  
  real(dp) :: ryd_to_ev, bohr_to_angstrom
  parameter ( ryd_to_ev = 13.60570444948592d0 )
  parameter ( bohr_to_angstrom = 0.5291772108d0 )
  
  !
  ! internal variables: end.
  !
  ! =============================================================================================
  

  contains
    
    SUBROUTINE QO_readin()
      USE io_files,      ONLY : nd_nmbr, prefix, tmp_dir, trimcheck
      USE io_global,     ONLY : ionode, stdout
      USE us,            ONLY : spline_ps
      IMPLICIT NONE
      INTEGER :: ios, ii
      character (len=256) :: outdir
      character(len=1), external :: capital
      
      NAMELIST / inputqo / &
           outdir, &
           prefix, &
           QO_E_window_max, &
           QO_E_window_min, &
           QO_flag_construction, &
           QO_flag_H_diagonal, &
           QO_flag_diagonalize_Hamiltonian, &
           QO_flag_orthonormalized, &
           QO_flag_normalized, &
           QO_flag_output_orbital, &           
           QO_flag_output_orbital_format, &
           QO_flag_output_orbital_rcutoff, &
           QO_unboundstate_cutoff, &
           QO_transport_direction
      
      
      ! 
      !   set default values for variables in namelist 
      ! 
      prefix = 'pwscf' 
      CALL get_env( 'ESPRESSO_TMPDIR', outdir )
      IF ( TRIM( outdir ) == ' ' ) outdir = './'
      
      QO_E_window_max = 0.d0
      QO_E_window_min = -1000.0d0
      QO_flag_construction = .true.
      QO_flag_H_diagonal = .false.
      QO_flag_diagonalize_Hamiltonian = .true.
      QO_unboundstate_cutoff = 0.d0
      QO_flag_orthonormalized = .false.
      QO_flag_normalized = .true.
      QO_transport_direction = 0
      
      if ( .not. ionode ) goto 400
      
      CALL input_from_file ( )
      
      READ (5, inputqo, err = 200, iostat = ios) 
200   CALL errore ('QO_readin', 'reading inputqo namelist', ABS (ios) )
      tmp_dir = trimcheck (outdir) 
      
      write(stdout, '(A)') ''
      write(stdout, '(A)') '----------------------------------------------------------------------'
      write(stdout, '(A)')  '[QO] input report: start ...'
      write(stdout, '(A)') ''
      write(stdout, *)     'outdir                  = ', trim(outdir)
      write(stdout, *)     'prefix                  = ', trim(prefix)
      if (QO_flag_construction) then
         write(stdout, *)  'QO_flag_construction    = .true.'
      else
         write(stdout, *)  'QO_flag_construction    = .false.'
      end if
      if ( QO_flag_H_diagonal ) then
         write(stdout, *)  'QO_flag_H_diagonal      = .true.'
      else
         write(stdout, *)  'QO_flag_H_diagonal      = .false.'
      end if
      if ( QO_flag_diagonalize_Hamiltonian ) then
         write(stdout, *)  'QO_flag_diagonalize_Hamiltonian      = .true.'
      else
         write(stdout, *)  'QO_flag_diagonalize_Hamiltonian      = .false.'
      end if
      write(stdout, *)     'QO_E_window_max         = ', QO_E_window_max
      write(stdout, *)     'QO_E_window_min         = ', QO_E_window_min
      write(stdout, *)     'QO_unboundstate_cutoff  = ', QO_unboundstate_cutoff
      
      if ( QO_flag_orthonormalized ) then
         write(stdout, *)  'QO_flag_orthonormalized = .true.'
      else
         write(stdout, *)  'QO_flag_orthonormalized = .false.'
      end if
      if ( QO_flag_normalized ) then
         write(stdout, *)  'QO_flag_normalized      = .true.'
      else
         write(stdout, *)  'QO_flag_normalized      = .false.'
      end if
      if (QO_flag_output_orbital) then
         write (stdout, *)    'QO_flag_output_orbital  = .true.'
         do ii = 1, len_trim( QO_flag_output_orbital_format )
            QO_flag_output_orbital_format( ii : ii ) = capital( QO_flag_output_orbital_format( ii : ii ) )
         end do
         if ( (trim(QO_flag_output_orbital_format) == 'XSF') .OR. &
              (trim(QO_flag_output_orbital_format) == 'NETCDF') .OR. &
              (trim(QO_flag_output_orbital_format) == 'NETCDFSMALL') .OR. &
              (trim(QO_flag_output_orbital_format) == 'BOTH') ) then
            write(stdout, *)  'QO_flag_output_orbital_format  = ', trim(QO_flag_output_orbital_format)
            write(stdout, *)  'QO_flag_output_orbital_rcutoff = ', QO_flag_output_orbital_rcutoff
         else
            call errore ('QO_readin', 'QO_flag_output_orbital_format is not recognized.', 10 )
         end if
      else
         write (stdout, *)    'QO_flag_output_orbital  = .false.'
      end if
      write(stdout, *)    'QO_transport_direction  = ', QO_transport_direction
      write(stdout, '(A)') ''
      write(stdout, '(A)')  '[QO] input report: end.'
      write(stdout, '(A)') '----------------------------------------------------------------------'
      write(stdout, *) ''
      
400   continue
      
      call QO_bcast_input
      
    END SUBROUTINE QO_readin
    
    !-----------------------------------------------------------------------
    ! Broadcast input data to all processors 
    !-----------------------------------------------------------------------
    SUBROUTINE QO_bcast_input
#ifdef __PARA
#include "f_defs.h"
      USE mp,            ONLY : mp_bcast
      USE io_files,      ONLY : prefix, tmp_dir
      USE us,            ONLY : spline_ps
      implicit none
      integer :: root = 0
      CALL mp_bcast( tmp_dir, root ) 
      CALL mp_bcast( prefix,  root ) 
      CALL mp_bcast( QO_E_window_max,  root )
      CALL mp_bcast( QO_E_window_min,  root )
      CALL mp_bcast( QO_flag_construction, root )
      CALL mp_bcast( QO_flag_H_diagonal,  root )
      CALL mp_bcast( QO_flag_diagonalize_Hamiltonian, root )
      CALL mp_bcast( QO_unboundstate_cutoff, root )
      CALL mp_bcast( QO_flag_orthonormalized, root )
      CALL mp_bcast( QO_flag_normalized, root )
      CALL mp_bcast( QO_flag_output_orbital, root )
      CALL mp_bcast( QO_flag_output_orbital_format, root )
      CALL mp_bcast( QO_flag_output_orbital_rcutoff, root )
      CALL mp_bcast( QO_transport_direction, root )
#endif
  END SUBROUTINE QO_bcast_input



subroutine QO_initialize_pp  
  use ions_base,  only : nat, ntyp => nsp, ityp
  use atom,       only : msh
  use us,         only : nqx
  implicit none
  integer     :: ii
  
  allocate( QO_PseudoatomicOrbital_nlm (ntyp, nchix_max) )
  allocate( QO_PseudoatomicOrbital_rescalingfactor (ntyp, nchix_max) )
  allocate( QO_PseudoatomicOrbital_nlm_effective(ntyp) )
  allocate( QO_PseudoatomicOrbital_channel_add(ntyp, nchix_max) )
  allocate( QO_PseudoatomicOrbital_channel_remove(ntyp, nchix_max) )
  
  QO_PseudoatomicOrbital_nlm(:,:)             = - 1
  QO_PseudoatomicOrbital_nlm_effective(:)     = 0     
  QO_PseudoatomicOrbital_rescalingfactor(:,:) = 0.d0
  QO_PseudoatomicOrbital_channel_add(:,:)     = - 1
  QO_PseudoatomicOrbital_channel_remove(:,:)  = 0
  
  QO_plot_natom = nat
  !!   ndm = MAXVAL(msh(1:ntyp))
  allocate (tab_at_lm (nqx, nchix_max, ntyp)                  )
  allocate (QO_plot_atomindex( QO_plot_natom )                )
  allocate (QO_plot_atomflag ( nat )                          )
  allocate (QO_PseudoatomicOrbital_number( nat )              )
  do ii = 1, QO_plot_natom
     QO_plot_atomindex(ii) = ii
     QO_plot_atomflag(ii)  = 1
  end do
end subroutine QO_initialize_pp


subroutine QO_initialize_atomicwfc
  use kinds,      only : DP
  use io_global,  only : stdout, ionode
  use basis,      only : natomwfc
  use wvfct,      only : npwx, nbnd, btype
  use noncollin_module, only : noncolin, npol
  use lsda_mod,   only : nspin
  use io_files,   only : iunat, iunsat, nwordatwfc, find_free_unit
  use ldaU,       only : lda_plus_u, swfcatom
  use klist,      only : nks, nkstot
  use ktetra,     only : nk1, nk2, nk3, k1, k2, k3  
  use uspp_param, only : upf 
  use ions_base,  only : nat, ntyp => nsp, ityp
  use smooth_grid_dimensions, only : nr1s, nr2s, nr3s, nr1sx, nr2sx, nr3sx, nrxxs
  use gsmooth,    only : doublegrid, nls, nlsm
  use grid_dimensions, only: nr1, nr2, nr3, nr1x, nr2x, nr3x, nrxx
  use cell_base,  only : at
  use fft_base,   only : cscatter_smooth
  use atom,       only : msh
  use us,         only : nqx
  
  implicit none  
  logical     :: exst
  logical     :: opnd  
  integer     :: nt, nb, l, startq, ii, jj, i, j, k, icount, it, remove_flag
  integer     :: t_nwfc, m_nwfc, m_mesh
  integer,     allocatable :: t_lchi(:)
  real(DP),    allocatable :: t_oc(:), t_chi(:,:)
  
  ! remove existing channels
  do nt = 1, ntyp
     
     m_nwfc = nchix_max
     m_mesh = msh(nt)
     
     if (ionode) write(stdout, *) 'nt, channel_remove = ', nt, QO_PseudoatomicOrbital_channel_remove(nt, :)
     
     allocate( t_oc  (m_nwfc) )
     allocate( t_lchi(m_nwfc) )
     allocate( t_chi (m_mesh, m_nwfc) )
     t_oc(:)     = 0.d0
     t_lchi(:)   = -1
     t_chi(:,:)  = 0.d0
     
     ii = 0
     do nb = 1, upf(nt)%nwfc
        remove_flag = 0
        do jj = 1, upf(nt)%nwfc
           if ( nb == QO_PseudoatomicOrbital_channel_remove(nt, jj) ) then
              remove_flag = 1
           end if
        end do
        if (remove_flag == 0) then
           ii           = ii + 1
           t_nwfc       = ii
           ! t_nwfc = upf(nt)%nwfc
           t_oc(ii)   = upf(nt)%oc(nb)
           t_lchi(ii) = upf(nt)%lchi(nb)
           t_chi(1:msh(nt), ii) = upf(nt)%chi(1:msh(nt), nb)  
        end if
     end do
     
     if (ionode) write (*,*) 't_nwfc = ', t_nwfc
     
     deallocate( upf(nt)%oc   )
     deallocate( upf(nt)%lchi )
     deallocate( upf(nt)%chi  )
     
     allocate(upf(nt)%oc   (t_nwfc))
     allocate(upf(nt)%lchi (t_nwfc))
     allocate(upf(nt)%chi  (m_mesh, t_nwfc))
     upf(nt)%nwfc            = t_nwfc
     upf(nt)%oc(1:t_nwfc)    = t_oc(1:t_nwfc)
     upf(nt)%lchi(1:t_nwfc)  = t_lchi(1:t_nwfc)
     upf(nt)%chi(1:m_mesh, 1:t_nwfc) = t_chi(1:m_mesh, 1:t_nwfc)     
     deallocate( t_oc, t_chi, t_lchi )
     
  end do



  ! add additional channels
  do nt = 1, ntyp
     
     m_nwfc = nchix_max
     m_mesh = msh(nt)
     
     ! if (ionode) write(stdout, *) 'nt, msh = ', nt, msh(nt)
     
     allocate( t_oc  (m_nwfc) )
     allocate( t_lchi(m_nwfc) )
     allocate( t_chi (m_mesh, m_nwfc) )
     t_oc(:)     = 0.d0
     t_lchi(:)   = -1
     t_chi(:,:)  = 0.d0
     
     t_nwfc = upf(nt)%nwfc
     t_oc(1:upf(nt)%nwfc)   = upf(nt)%oc(1:upf(nt)%nwfc)
     t_lchi(1:upf(nt)%nwfc) = upf(nt)%lchi(1:upf(nt)%nwfc)
     t_chi(1:msh(nt), 1:upf(nt)%nwfc) = upf(nt)%chi(1:msh(nt), 1:upf(nt)%nwfc)
     
     deallocate( upf(nt)%oc   )
     deallocate( upf(nt)%lchi )
     deallocate( upf(nt)%chi  )
     
     ii = t_nwfc
     do nb = 1, nchix_max
        if (QO_PseudoatomicOrbital_channel_add(nt, nb) >= 0) then           
           ii = ii + 1
           t_nwfc       = t_nwfc + 1
           t_oc(ii)     = 0.d0
           t_lchi(ii)   = QO_PseudoatomicOrbital_channel_add(nt, nb) 
           t_chi(1:msh(nt), ii) = t_chi(1:msh(nt), ii-1)
        end if
     end do
     
     allocate(upf(nt)%oc   (t_nwfc))
     allocate(upf(nt)%lchi (t_nwfc))
     allocate(upf(nt)%chi  (m_mesh, t_nwfc))
     upf(nt)%nwfc            = t_nwfc
     upf(nt)%oc(1:t_nwfc)    = t_oc(1:t_nwfc)
     upf(nt)%lchi(1:t_nwfc)  = t_lchi(1:t_nwfc)
     upf(nt)%chi(1:m_mesh, 1:t_nwfc) = t_chi(1:m_mesh, 1:t_nwfc)     
     deallocate( t_oc, t_chi, t_lchi )
     
  end do

  natomwfc = n_atom_wfc(nat, ityp)
  
  write(stdout, *) 'natomwfc = ', natomwfc

end subroutine QO_initialize_atomicwfc

!----------------------------------------------------------------------------
FUNCTION n_atom_wfc( nat, ityp )
  !----------------------------------------------------------------------------
  !
  ! ... Find number of starting atomic orbitals
  !
  USE uspp_param,       ONLY : upf
  USE noncollin_module, ONLY : noncolin
  USE io_global,        ONLY : stdout
  !
  IMPLICIT NONE
  !
  INTEGER, INTENT(IN)  :: nat, ityp(nat)
  !
  INTEGER  :: n_atom_wfc
  !
  INTEGER  :: na, nt, n
  !
  !
  n_atom_wfc = 0
  !
  DO na = 1, nat
     !
     nt = ityp(na)
     !
     DO n = 1, upf(nt)%nwfc
        !
        IF ( upf(nt)%oc(n) >= 0.D0 ) THEN
           !
           IF ( noncolin ) THEN
              !
              IF ( upf(nt)%has_so ) THEN
                 !
                 n_atom_wfc = n_atom_wfc + 2 * upf(nt)%lchi(n)
                 !
                 ! write(stdout, *) 'info: ', nt, n, upf(nt)%jchi(n), upf(nt)%lchi(n), upf(nt)%nwfc
                 IF ( ABS( upf(nt)%jchi(n)-upf(nt)%lchi(n) - 0.5D0 ) < 1.D-6 ) &
                    n_atom_wfc = n_atom_wfc + 2
                 !
              ELSE
                 !
                 n_atom_wfc = n_atom_wfc + 2 * ( 2 * upf(nt)%lchi(n) + 1 )
                 !
              END IF
              !
           ELSE
              !
              n_atom_wfc = n_atom_wfc + 2 * upf(nt)%lchi(n) + 1
              !
           END IF
        END IF
     END DO
  END DO
  !
  RETURN
  !
END FUNCTION n_atom_wfc
!

!----------------------------------------------------------------------------
FUNCTION n_element_atomic_orbital( ityp )
  !----------------------------------------------------------------------------
  !
  ! ... Find number of atomic orbitals on element ityp
  !
  USE uspp_param,       ONLY : upf
  USE noncollin_module, ONLY : noncolin
  USE io_global,        ONLY : stdout
  !
  IMPLICIT NONE
  !
  INTEGER, INTENT(IN)  :: ityp
  !
  INTEGER  :: n_element_atomic_orbital
  !
  INTEGER  :: nt, n
  !
  !
  n_element_atomic_orbital = 0
  !
  nt = ityp
  !
  DO n = 1, upf(nt)%nwfc
     !
     IF ( upf(nt)%oc(n) >= 0.D0 ) THEN
        !
        IF ( noncolin ) THEN
           !
           IF ( upf(nt)%has_so ) THEN
              !
              n_element_atomic_orbital = n_element_atomic_orbital + 2 * upf(nt)%lchi(n)
              !
              ! write(stdout, *) 'info: ', nt, n, upf(nt)%jchi(n), upf(nt)%lchi(n), upf(nt)%nwfc
              IF ( ABS( upf(nt)%jchi(n)-upf(nt)%lchi(n) - 0.5D0 ) < 1.D-6 ) &
                   n_element_atomic_orbital = n_element_atomic_orbital + 2
              !
           ELSE
              !
              n_element_atomic_orbital = n_element_atomic_orbital + 2 * ( 2 * upf(nt)%lchi(n) + 1 )
              !
           END IF
           !
        ELSE
           !
           n_element_atomic_orbital = n_element_atomic_orbital + 2 * upf(nt)%lchi(n) + 1
           !
        END IF
     END IF
  END DO
  !
  RETURN
  !
END FUNCTION n_element_atomic_orbital
!

subroutine QO_initialize()
  
  use kinds,      only : DP
  use io_global,  only : stdout, ionode
  use basis,      only : natomwfc
  use wvfct,      only : npwx, nbnd, btype
  use noncollin_module, only : noncolin, npol
  use lsda_mod,   only : nspin
  use io_files,   only : iunat, iunsat, nwordatwfc, find_free_unit
  use ldaU,       only : lda_plus_u, swfcatom
  use klist,      only : nks, nkstot
  use ktetra,     only : nk1, nk2, nk3, k1, k2, k3  
  use uspp_param, only : upf 
  use ions_base,  only : nat, ntyp => nsp, ityp
  use smooth_grid_dimensions, only : nr1s, nr2s, nr3s, nr1sx, nr2sx, nr3sx, nrxxs
  use gsmooth,    only : doublegrid, nls, nlsm
  use grid_dimensions, only: nr1, nr2, nr3, nr1x, nr2x, nr3x, nrxx
  use cell_base,  only : at
  use fft_base,   only : cscatter_smooth
  use atom,       only : msh
  use us,         only : nqx
  use io_files,   only : diropn
  
  implicit none  
  logical     :: exst
  logical     :: opnd  
  integer     :: nt, nb, l, startq, ii, i, j, k, icount, ic, it
  real(dp)    :: H(3,3)
  complex(dp), allocatable :: cXgrid(:,:), cXgrid_total(:,:)

  call start_clock ('initialize')
  
  if (ionode) then
     write(stdout, '(A)') ''
  end if
  
  if ( .not.  l_QO_initialized ) then
     
     if (ionode) then
        write(stdout, '(A)') '----------------------------------------------------------------------'
        write(stdout, '(A)') '[QO] initialize: start ...'
     end if
     
     do nt = 1, ntyp
        ii = 1
        do nb = 1, upf(nt)%nwfc
           if (upf(nt)%oc (nb) >= 0.d0) then
              QO_PseudoatomicOrbital_nlm( nt, ii) = upf(nt)%lchi(nb)
              ii = ii + 1
           end if
        end do
        QO_PseudoatomicOrbital_nlm_effective( nt ) = ii - 1
     end do
     
     if ( dble(nk1*nk2*nk3) > 0.d0) then
        ncellhalfx = floor(dble(nk1-1)/2.d0)
        ncellhalfy = floor(dble(nk2-1)/2.d0)
        ncellhalfz = floor(dble(nk3-1)/2.d0)
     else
        ncellhalfx = 0
        ncellhalfy = 0
        ncellhalfz = 0
     end if
     
     ncellx = 2 * ncellhalfx + 1
     ncelly = 2 * ncellhalfy + 1
     ncellz = 2 * ncellhalfz + 1
     ncell  = ncellx * ncelly * ncellz
     nkxyz  = nk1    * nk2    * nk3
     nrxx_total  = nr1x*nr2x*nr3x
     nrxxs_total = nr1sx*nr2sx*nr3sx
     
     nspin_general = nspin
     if (noncolin) nspin_general = 1
     
     allocate (QO_wfc        ( npwx*npol, natomwfc )             )
     allocate (QO_Omega      ( natomwfc, natomwfc, nkstot ) )
     allocate (QO_cmk        ( npwx*npol, natomwfc )             )
     allocate (QO_Enmk       ( natomwfc, natomwfc, nkstot )      )
     allocate (QO_OEO        ( natomwfc, natomwfc, nkstot)       )
     allocate (QO_OO         ( natomwfc, natomwfc, nkstot)       )
     allocate (QO_nbndk      ( nkstot )                          )
     allocate (QO_bandindex  ( nbnd, nkstot )                    )
     allocate (Bloch_projsum ( nbnd, nkstot )                    )
     allocate (AO_norm       ( natomwfc )                        )
     
     allocate (nbnd_occ      (nkstot)                            )
     
     allocate (QO_to_Atom    (natomwfc)                          )
     allocate (Atom_to_QO    (nat, 2)                            )
     allocate (QO_position   (3, natomwfc)                       )
     allocate (Atom_position (3, nat)                            )
     
     allocate (QO_wfc_r      ( nrxxs*npol, natomwfc, nspin_general )     )
     allocate (QO_H          ( ncell, natomwfc, natomwfc, nspin_general ))
     allocate (QO_charge     ( ncell, natomwfc, natomwfc, nspin_general ))
     allocate (QO_S          ( ncell, natomwfc, natomwfc, nspin_general ))
     
     QO_PseudoatomicOrbital_number = 0
     QO_wfc   = (0.0_DP, 0.0_DP)
     QO_wfc_r = (0.0_DP, 0.0_DP)
     QO_Omega = (0.0_DP, 0.0_DP)
     QO_cmk   = (0.0_DP, 0.0_DP)
     QO_Enmk  = (0.0_DP, 0.0_DP)
     QO_H     = (0.0_DP, 0.0_DP)
     QO_S     = (0.0_DP, 0.0_DP)
     QO_OEO   = (0.0_DP, 0.0_DP)
     QO_OO    = (0.0_DP, 0.0_DP)
     QO_charge= 0.0_DP
     QO_nbndk      = nbnd
     QO_bandindex  = 0
     Bloch_projsum = 0.d0
     AO_norm       = 0.d0

     do i = 1, nat
        QO_PseudoatomicOrbital_number(i) = 0
        it = ityp(i)
        QO_PseudoatomicOrbital_number(i) = n_element_atomic_orbital(it)
!        do ic = 1, nchix_max
!           if (QO_PseudoatomicOrbital_nlm(it, ic) >=0 ) then
!              QO_PseudoatomicOrbital_number(i) = QO_PseudoatomicOrbital_number(i) + &
!                   2 * QO_PseudoatomicOrbital_nlm(it, ic) + 1
!           end if
!        end do
     end do
     
     allocate( Xgrid (nrxxs,3)             )
     allocate( Xgrid_total (nrxxs_total,3) )
     allocate( cXgrid (nrxxs,3)            )
     allocate( cXgrid_total (nrxxs_total,3))

     write(*,*) 'QO_initialize: ', nr1sx, nr2sx, nr3sx, nrxxs, nrxxs_total
     icount = 1
     do k = 0, nr3sx-1
        do j = 0, nr2sx-1
           do i = 0, nr1sx-1
              Xgrid_total(icount, 1) =  dble(i)/dble(nr1sx)
              Xgrid_total(icount, 2) =  dble(j)/dble(nr2sx)
              Xgrid_total(icount, 3) =  dble(k)/dble(nr3sx)
              icount = icount + 1
           end do
        end do
     end do
     ! H for unitcell
     H(1:3, 1:3) = transpose(at(1:3,1:3))
     Xgrid_total = matmul(Xgrid_total, H)
     cXgrid_total = Xgrid_total
#ifdef __PARA
     call cscatter_smooth (cXgrid_total(:,1), cXgrid(:,1))
     call cscatter_smooth (cXgrid_total(:,2), cXgrid(:,2))
     call cscatter_smooth (cXgrid_total(:,3), cXgrid(:,3))
#endif
     Xgrid = dble(cXgrid)
     deallocate(cXgrid_total)
     deallocate(cXgrid)
     
!     if ( .not. lda_plus_u ) then
     nwordatwfc = 2 * natomwfc * npwx * npol
     call diropn( iunat,  'atwfc',  nwordatwfc, exst )
     call diropn( iunsat, 'satwfc', nwordatwfc, exst )
     
!    nwordatwfcr = 2 * natomwfc * nrxxs * npol
     nwordatwfcr = 2 * natomwfc * nrxxs * npol * nspin_general
     iunsatr = find_free_unit()
     call diropn( iunsatr, 'satwfcr', nwordatwfcr, exst )
     if (.not. allocated(swfcatom) ) allocate( swfcatom(npwx*npol, natomwfc) )
!     end if
     
     !! if (.not. allocated(btype)) allocate( btype(nbnd, nkstot) )
     
     l_QO_initialized     = .true.
     
     if (ionode) then
        write(stdout, '(A)') ''        
        write(stdout, '(A)') '[QO] initialize: end'
        write(stdout, '(A)') '----------------------------------------------------------------------'
        write(stdout, '(A)') ''
     end if
     
  else

     if (ionode) then
        write(stdout, '(A)') '[QO] it has been already initialized.'
     end if
     
  end if

  if (ionode) then
     write(stdout, '(A)') ''
  end if
  
  call stop_clock ('initialize')
  
end subroutine QO_initialize


subroutine QO_finalize

  use io_files,      only: iunat, iunsat, find_free_unit
  use ldaU,          only: lda_plus_u, swfcatom
  use io_global,     only: stdout, ionode
  use mp,            only : mp_barrier  
  
  logical :: opnd

  call start_clock ('finalize')
  
  if (ionode) then
     write(stdout, '(A)') ''
     write(stdout, '(A)') '----------------------------------------------------------------------'
     write(stdout, '(A)') '[QO] finalize: start ...'
     write(stdout, '(A)') ''
  end if
  
  if ( l_QO_initialized ) then
     if (allocated(QO_plot_kpoints))   deallocate (QO_plot_kpoints)
     if (allocated(QO_plot_kpoints_index))   deallocate (QO_plot_kpoints_index)
     if (allocated(QO_plot_atomindex)) deallocate (QO_plot_atomindex)
     if (allocated(QO_plot_atomflag))  deallocate (QO_plot_atomflag)
     if (allocated(tab_at_lm))    deallocate (tab_at_lm)
     if (allocated(QO_wfc)   )    deallocate( QO_wfc       )
     if (allocated(QO_wfc_r)   )  deallocate( QO_wfc_r     )
     if (allocated(QO_Omega) )    deallocate( QO_Omega     )
     if (allocated(QO_cmk)   )    deallocate( QO_cmk       )
     if (allocated(QO_Enmk)  )    deallocate( QO_Enmk      )
     if (allocated(QO_H)     )    deallocate( QO_H         )
     if (allocated(QO_S)     )    deallocate( QO_S         )
     if (allocated(QO_OEO)   )    deallocate( QO_OEO       )
     if (allocated(QO_OO)    )    deallocate( QO_OO        )
     if (allocated(QO_charge))    deallocate( QO_charge    )
     if (allocated(QO_nbndk) )    deallocate( QO_nbndk     )
     if (allocated(QO_bandindex)) deallocate( QO_bandindex )
     if (allocated(QO_PseudoatomicOrbital_nlm))              deallocate( QO_PseudoatomicOrbital_nlm )
     if (allocated(QO_PseudoatomicOrbital_rescalingfactor))  deallocate( QO_PseudoatomicOrbital_rescalingfactor )
     if (allocated(QO_PseudoatomicOrbital_nlm_effective))    deallocate( QO_PseudoatomicOrbital_nlm_effective )
     if (allocated(QO_PseudoatomicOrbital_channel_add) )     deallocate( QO_PseudoatomicOrbital_channel_add )
     if (allocated(QO_PseudoatomicOrbital_channel_remove) )  deallocate( QO_PseudoatomicOrbital_channel_remove )
     if (allocated(QO_PseudoatomicOrbital_number) )          deallocate( QO_PseudoatomicOrbital_number )
     if (allocated(Bloch_projsum))                           deallocate( Bloch_projsum )
     if (allocated(AO_norm))                                 deallocate( AO_norm )
     if (allocated(nbnd_occ))                                deallocate( nbnd_occ )
     if (allocated(swfcatom))                                deallocate( swfcatom )
     
     if (allocated(QO_to_Atom))    deallocate (QO_to_Atom    )
     if (allocated(Atom_to_QO))    deallocate (Atom_to_QO    )
     if (allocated(QO_position))   deallocate (QO_position   )
     if (allocated(Atom_position)) deallocate (Atom_position )
     if (allocated(Xgrid))         deallocate (Xgrid         )
     if (allocated(Xgrid_total))   deallocate (Xgrid_total   )
     
     inquire( unit = iunat, opened = opnd )
     if ( opnd ) close( unit = iunat,  status = 'keep' )
     inquire( unit = iunsat, opened = opnd )
     if ( opnd ) close( unit = iunsat, status = 'keep' )
        
     l_QO_initialized     = .false.
     
  else
     
     if (ionode) then
        write(stdout, '(A)') ' MODULE QO was NOT initialized at all.'
        write(stdout, '(A)') ''
     end if
     
  end if
  
  if (ionode) then
     write(stdout, '(A)') '[QO] finalize: end.'
     write(stdout, '(A)') '----------------------------------------------------------------------'
     write(stdout, '(A)') ''
  end if
  
  call stop_clock ('finalize')
    
end subroutine QO_finalize



subroutine QO_setup_dft
  
  use uspp,          only : nkb, vkb
  use ions_base,     only : ntyp => nsp
  use dfunct,        only : newd
  use wvfct,         only : npwx, npw, igk, nbnd, et, wg, g2kin, current_k
  use uspp_param,    only : upf
  use scf,           only : v, vrs, vltot, rho, rho_core, kedtau, rhog_core
  use grid_dimensions,  only : nr1, nr2, nr3, nrxx
  use gvect,         only : g, ngm, ecutwfc
  use klist,         only : nkstot, nks, xk, ngk, wk
  use cell_base,     only : omega, alat, at, bg, tpiba, tpiba2
  use io_files,      only : nwordwfc, nwordatwfc, iunwfc, iunigk, iunat, iunsat
  use buffers,       only : get_buffer
  use wavefunctions_module, only : evc
  use io_global,     only : stdout, ionode
  use smooth_grid_dimensions, only : nr1s, nr2s, nr3s, nr1sx, nr2sx, nr3sx, nrxxs
  use gsmooth,       only : doublegrid, nls, nlsm
  use lsda_mod,      only : nspin, lsda, isk, current_spin
#if defined (EXX) 
  use exx,           only : exx_grid_init, exx_div_check, exx_divergence, &
                            exxdiv, erfc_scrlen, exxinit
  use funct,         only : dft_is_hybrid, exx_is_active, start_exx, &
                            get_exx_fraction, get_screening_parameter, &
                            write_dft_name
#endif
  
  real(dp) :: charge
  integer  :: ik
  
  any_uspp = ANY(upf(1:ntyp)%tvanp)
  
  if (nkb > 0) then
     call init_us_1 
  end if
  call QO_init_at_1
  
  ! computes the total local potential (external+scf) on the smooth grid
  call setlocal
  
  ! prepare igk for exx
  rewind (iunigk)
  do ik = 1, nks
     current_k = ik
     npw = ngk (ik)
     call get_buffer (evc, nwordwfc, iunwfc, ik)
     call gk_sort (xk (1, ik), ngm, g, ecutwfc / tpiba2, npw, igk, g2kin)
     g2kin(:) = g2kin(:) * tpiba2
     if ( nks > 1 ) write( iunigk ) igk
     if ( nkb > 0 ) then
        ! compute pseudopotential projectors
        call init_us_2 (npw, igk, xk(1, ik), vkb)
     end if
  end do
  
#if defined (EXX)
  IF ( dft_is_hybrid() ) THEN
     write(stdout,  *) '--------------------------------------------------------------------'
     call write_dft_name()
     erfc_scrlen = get_screening_parameter()
     exxdiv = exx_divergence()
     call exxinit
     if ( exx_is_active() ) write(stdout, *) '  exx is active.'
     write(stdout,  *) '--------------------------------------------------------------------'
  ENDIF
#endif

  !! calculate total charge density
  !call sum_band()
  
  !! LDA+U
  !if (lda_plus_U) then
  !   call write_ns()
  !end if
  
  ! calculate HXC-potential
  call v_of_rho( rho, rho_core, rhog_core, &
       ehart, etxc, vtxc, eth, etotefield, charge, v )
  
  ! calculate total local potential (external + scf)
  call set_vrs( vrs, vltot, v%of_r, kedtau, v%kin_r, nrxx, nspin, doublegrid )
  
  ! compute the D_{nm}^{I} matrix for the nonlocal pseudopotentials
  call newd()
  
end subroutine QO_setup_dft


subroutine QO_construction
  
  use kinds,      only : dp
  use io_global,  only : stdout, ionode
  use io_files,   only : nwordwfc, nwordatwfc, iunwfc, iunigk, iunat, iunsat
  use ions_base,  only : nat
  use basis,      only : natomwfc
  use klist,      only : nkstot, nks, xk, ngk, wk
  use ldaU,       only : swfcatom, U_projection, eth
  use wvfct,      only : npwx, npw, igk, nbnd, et, wg, g2kin, current_k
  use uspp,       only : nkb, vkb
  use noncollin_module, only : noncolin, npol
  use lsda_mod,         only : nspin, lsda, isk, current_spin
  use wavefunctions_module, only : evc
  use buffers,              only : get_buffer
  use gvect,     only : g, ngm, ecutwfc
  use cell_base, only : omega, alat, at, bg, tpiba, tpiba2
  use smooth_grid_dimensions, only : nr1s, nr2s, nr3s, nr1sx, nr2sx, nr3sx, nrxxs
  use gsmooth,   only : doublegrid, nls, nlsm
  use symm_base,     only : invsym
  use ener
  use ions_base,     only : ntyp => nsp
  use control_flags, only : gamma_only
  use mp_global,     only : inter_pool_comm, intra_pool_comm 
  use mp,            only : mp_sum
  use constants,     only : tpi
  use ldaU,          only : lda_plus_u, swfcatom  
  USE fft_base,       ONLY : dffts
  USE fft_interfaces, ONLY : invfft, fwfft
  
  implicit none
  
  integer :: ik, ibnd, info, i, j, k, na, nb, nt, isym, n, ntemp, m, &
       l, lm, ltot, ntot, ipol, ispin, id, itruebnd, istart, iend, ip, ii, n_cmk
  logical :: exst
  
  complex(dp) :: temp, t (5), temp_enk
  complex(dp),  allocatable :: wfcatom(:,:), work(:,:), overlap(:,:),  overlap2(:,:), Omega_old(:,:)
  real(dp),     allocatable :: e (:)
  complex(dp),  allocatable :: psic(:), phase(:)  
  complex(dp) , allocatable :: wfcatom_para(:,:)
  complex(dp) , allocatable :: wfcatom_perp(:,:)
  complex(dp) , allocatable :: proj(:,:)
  real(dp),     allocatable :: projsum(:)
  complex(dp) , allocatable :: W(:,:)
  complex(dp) , allocatable :: wfc_temp(:), wfc_temp2(:,:)
  complex(dp) , allocatable :: Enmk_temp(:,:)
  complex(dp) , allocatable :: Enmk_c(:,:), work_c(:,:)
  real(dp),     allocatable :: e_c(:)
  complex(dp) , external    :: zdotc
  
  logical :: opnd, any_uspp
  real(dp) :: ek, eloc, enl, charge, etotefield, wk_tot
  real(dp) :: xkprime(3)
  real(dp) :: tmp_real
  real(dp), allocatable :: AO_norm_k(:)
  real(dp) :: wfcnorm
  real(dp) :: ryd_to_ev
  parameter ( ryd_to_ev = 13.60570444948592d0 )
  
  
  call start_clock ('construction')
  
  if (ionode) then
     write(stdout, '(A)') ''
     write(stdout, '(A)') '----------------------------------------------------------------------'
     write(stdout, '(A)') '[QO] QO construction: start ...' 
  end if
  
  allocate (wfcatom( npwx*npol, natomwfc)     )
  allocate (wfcatom_para( npwx*npol, natomwfc))
  allocate (wfcatom_perp( npwx*npol, natomwfc))
  allocate( proj(nbnd, natomwfc) )
  allocate( projsum(nbnd)        )
  allocate( W(natomwfc,natomwfc) )
  
  allocate (psic (nrxxs)                   )
  allocate (phase(nrxxs)                   )
  allocate (Omega_old(natomwfc,natomwfc)   )
  allocate (work   ( natomwfc , natomwfc)  )
  allocate (overlap( natomwfc , natomwfc)  )
  allocate (overlap2( natomwfc , natomwfc) )
  allocate (e      ( natomwfc)             )    
  allocate (wfc_temp(npwx*npol)            )
  allocate (wfc_temp2(npwx*npol, 1)        )
  allocate (Enmk_temp( natomwfc , natomwfc))
  allocate (AO_norm_k( natomwfc )          )    
  
  ! calculate AO_norm for the normalization in the next step
  AO_norm   = 0.d0
  wk_tot    = 0.d0  
  if (nks > 1) rewind (iunigk)
  do ik = 1, nks     
     
     current_k = ik
     if (lsda) current_spin = isk(ik)
     if (noncolin) current_spin = 1
     
     ! read g_vectors and bloch wavefunctions
     call gk_sort (xk (1, ik), ngm, g, ecutwfc / tpiba2, npw, igk, g2kin)
     g2kin(:) = g2kin(:) * tpiba2
     
     npw = ngk (ik)
     ! call davcio (evc, nwordwfc, iunwfc, ik, -1) 
     call get_buffer (evc, nwordwfc, iunwfc, ik)
     
     if (nkb > 0) then
        ! compute pseudopotential projectors
        call init_us_2 (npw, igk, xk(1, ik), vkb)
     end if
     
     wfcatom = cmplx(0.d0, 0.d0)
     ! compute atomic wavefunctions
     call QO_init_atomic_wfc (ik, wfcatom)
     
     ! calculate <AO_k|S|AO_k>
     AO_norm_k = 0.d0
     do j = 1, natomwfc
        wfc_temp(:) = cmplx(0.d0, 0.d0)
        call QO_s_psi (npwx, npw, 1, wfcatom(1, j), wfc_temp)
        AO_norm_k(j) = dble( ZDOTC (npwx*npol,  wfcatom(1, j), 1, wfc_temp, 1) )
     end do
     
#ifdef __PARA
     ! recover over G parallelization (intra_pool)
     call mp_sum(AO_norm_k, intra_pool_comm)
#endif
     
     AO_norm(1:natomwfc) = AO_norm(1:natomwfc) + AO_norm_k(1:natomwfc) * wk(ik)
     wk_tot = wk_tot + wk(ik)
  end do
  
#ifdef __PARA
  ! recover over kpt parallelization (inter_pool)
  call mp_sum (AO_norm, inter_pool_comm)
  call mp_sum (wk_tot,  inter_pool_comm)
#endif
  
  
  ! if (invsym)   AO_norm = AO_norm / 2.d0   ! [TAG] need to check the noncollinear case
  AO_norm = AO_norm / wk_tot
  
  if (ionode) then
     write(stdout, '(A)') ''
     write(stdout, '(4x, A, f16.12)') 'wk_tot = ', wk_tot
     write(stdout, '(4x, A)') 'norm of AO -- <AO|S|AO> (before normalization): '
     do i = 1, natomwfc
        write(stdout, '(4x, i6, f16.12)') i, AO_norm(i)
     end do
     write(stdout, '(A)') ''
  end if
  
  if (nks > 1) rewind (iunigk)
  do ik = 1, nks
     
     write(stdout, '(A)') ''
     write(stdout, '(4x, A)') '------------------------------------------------------------------'
     
     i = 1
     do while ( i<=nbnd )
        if ((et(i, ik) - ef) * ryd_to_ev <=  (QO_E_window_max + 1.0d-6) ) then
           QO_nbndk(ik) = i
        end if
        i = i + 1
     end do
     
     current_k = ik
     if (lsda) current_spin = isk(ik)
     if (noncolin) current_spin = 1
        
     ! read g_vectors and bloch wavefunctions
     call gk_sort (xk (1, ik), ngm, g, ecutwfc / tpiba2, npw, igk, g2kin) 
     g2kin(:) = g2kin(:) * tpiba2
     
     npw = ngk (ik)
     ! call davcio (evc, nwordwfc, iunwfc, ik, -1) 
     call get_buffer (evc, nwordwfc, iunwfc, ik)
     
     if (nkb > 0) then
        ! compute pseudopotential projectors
        call init_us_2 (npw, igk, xk(1, ik), vkb)
     end if
     
     wfcatom = cmplx(0.d0, 0.d0)
     ! compute atomic wavefunctions
     call QO_init_atomic_wfc (ik, wfcatom)
     
     do i = 1, natomwfc
        wfcatom(:,i) = wfcatom(:,i) / dsqrt(AO_norm(i))
     end do
     
     ! write atomic wfc on unit iunat
     call davcio (wfcatom, nwordatwfc, iunat, ik, 1)
     
     ! compute <psi_nk(occupied)|S|AO> 
     proj = cmplx(0.d0, 0.d0)     
     swfcatom = cmplx(0.d0, 0.d0)
     do i = 1, natomwfc
        wfc_temp(:) = cmplx(0.d0, 0.d0)
        call QO_s_psi (npwx, npw, 1, wfcatom(1, i), wfc_temp)
        swfcatom(:, i) = wfc_temp(:)
        do ibnd = 1, QO_nbndk(ik)
           proj(ibnd, i) = ZDOTC(npwx*npol, evc (1, ibnd), 1, &
                swfcatom (1, i), 1)
        end do
     end do     
#ifdef __PARA
     ! recover over G parallelization (intra_pool)
     call mp_sum (proj, intra_pool_comm)
#endif
     
     
     ! remove the unbound state
     projsum = 0.d0
     id = 0
     QO_bandindex(1:nbnd, ik) = 0
     if (QO_unboundstate_cutoff > 1.d-8) then
        
        do ibnd = 1, QO_nbndk(ik)
           do i = 1, natomwfc
              projsum(ibnd) = projsum(ibnd) + &
                   dble( proj(ibnd, i) * conjg(proj(ibnd, i)) )
           end do
           
           if ( (id < nbnd) .and. (id < natomwfc) ) then
              if ( projsum(ibnd) >= QO_unboundstate_cutoff ) then
                 id = id + 1
                 QO_bandindex(id, ik) = ibnd
              end if
           end if           
        end do
        
     else
        
        do ibnd = 1, QO_nbndk(ik)
           do i = 1, natomwfc
              projsum(ibnd) = projsum(ibnd) + &
                   dble( proj(ibnd, i) * conjg(proj(ibnd, i)))
           end do
           QO_bandindex(ibnd, ik) = ibnd
        end do
        id = QO_nbndk(ik)
        
     end if
     
     QO_nbndk(ik) = id
     Bloch_projsum(:, ik) = projsum(:)
     
     write(stdout, '(4x, A, 3i8, i16)') 'ik, QO_nbndk(ik), natomwfc, nwordatwfc = ', ik, QO_nbndk(ik), natomwfc, nwordatwfc
     write(stdout, '(A)') ''
     
     write(stdout, '(8x, A, i4, A, 3f10.6)') 'xk(', ik, ') = ', xk(1, ik), xk(2, ik), xk(3, ik)
     write(stdout, '(A)') ''
     
     
     ! compute  |AO^{\parallel}> = \sum_{n} |psi_nk> <psi_nk|S|AO> (for occupied Bloch states)
     wfcatom_para = cmplx(0.d0, 0.d0)
     do i = 1, natomwfc
        do ibnd = 1, QO_nbndk(ik)
           itruebnd = QO_bandindex(ibnd, ik)
           wfcatom_para(:, i) = wfcatom_para(:, i) + &
                evc(:,itruebnd) * proj(itruebnd, i)
        end do
     end do
     
     ! compute  |AO^{\perp}> =  |AO> - |AO^{\parallel}> 
     wfcatom_perp = cmplx(0.d0, 0.d0)
     do i = 1, natomwfc
        wfcatom_perp(:, i) = wfcatom(:, i) - wfcatom_para(:, i) 
     end do
     
     ! compute W(i,j)^{ik}  = <wfcatom_perp(i, ik) |  S| wfcatom_perp(j, ik)>
     W = cmplx(0.d0, 0.d0)
     do j = 1, natomwfc
        wfc_temp(:) = cmplx(0.d0, 0.d0)
        call QO_s_psi (npwx, npw, 1, wfcatom_perp(1, j), wfc_temp)
        do i = j, natomwfc
           W(i, j) = ZDOTC (npwx*npol,  wfcatom_perp(1, i), 1, wfc_temp, 1)
           W(j, i) = conjg(W(i, j))
           if (i == j ) then
              W(i, i) = cmplx( dble(W (i, i)), 0.d0 )
           end if
        end do
     end do
#ifdef __PARA
     ! recover over G parallelization (intra_pool)
     call mp_sum ( W, intra_pool_comm )
#endif
     
     ! diagonalize W(i,j) matrix 
     work(:,:) = cmplx(0.d0,0.d0)
     e(:) = 0.d0
     call cdiagh (natomwfc, W, natomwfc, e, work)
     
     ! choose the largest eigenvalues and the corresponding eigenvectors
     write(stdout, '(8x, A)') 'eigenvalues of W matrix'
     do ibnd = 1, natomwfc
        write(stdout, '(12x, 2i6, f10.6)') ik, ibnd, e(ibnd)
     end do
     write(stdout, '(A)') ''
     
     ! construct |c_mk> below the frozen window (the same as Kohn-Sham eigenstates)
     QO_cmk(:,:) = cmplx(0.d0, 0.d0)
     do ibnd = 1, QO_nbndk(ik)
        itruebnd = QO_bandindex(ibnd, ik)
        QO_cmk(:,ibnd) = evc(:, itruebnd)
     end do
     
     ! construct |c_mk> above the frozen window (optimized combination unoccupied subspace)
     do i = 1, (natomwfc - QO_nbndk(ik))
        do j = 1, natomwfc
           QO_cmk(:, QO_nbndk(ik)+i) = QO_cmk(:, QO_nbndk(ik)+i) + &
                1.d0 / dsqrt( e(natomwfc-i+1) ) * &
                work(j, natomwfc-i+1) * wfcatom_perp(:, j)
        end do
     end do
     
     !
     ! overlap_bloch_qo =  (occupied):    overlap_bloch_and_ao(n,j)
     !                     (unoccupied):  conj(work(j,n) * dsqrt(e(n)))
     QO_Omega( :, :, ik ) = cmplx(0.d0, 0.d0)
     do i = 1, natomwfc
        do ibnd = 1, QO_nbndk(ik)
           itruebnd = QO_bandindex(ibnd, ik)
           QO_Omega( ibnd, i, ik ) = proj(itruebnd, i)
        end do
        do ibnd = 1, (natomwfc - QO_nbndk(ik))
           QO_Omega( QO_nbndk(ik)+ibnd, i, ik ) = conjg( work(i, natomwfc-ibnd+1) * dsqrt(e(natomwfc-ibnd+1)) )
        end do
     end do
     
     ! compute <c_mk| H |c_nk> for optimized manifold
     Enmk_temp = cmplx(0.d0, 0.d0)
     do i = ( QO_nbndk(ik)+1 ), natomwfc
        wfc_temp(:) = cmplx(0.d0, 0.d0)
        call QO_h_psi (npwx, npw, 1, QO_cmk(1, i), wfc_temp)
        
        if ( QO_flag_H_diagonal ) then
           Enmk_temp(i, i) = &
                cmplx( dble(ZDOTC (npwx*npol, QO_cmk(1, i) , 1, wfc_temp, 1)), 0.d0)
        else
           do j = i, natomwfc
              Enmk_temp(j, i) = ZDOTC (npwx*npol, QO_cmk(1, j) , 1, wfc_temp, 1)
              Enmk_temp(i, j) = conjg( Enmk_temp(j, i) )
              if (i == j) then
                 Enmk_temp(i, i) = cmplx(dble(Enmk_temp(i, i)), 0.d0)
              end if
           end do
        end if
     end do
     
#ifdef __PARA
     ! recover over G parallelization (intra_pool)
     call mp_sum (Enmk_temp, intra_pool_comm)
#endif
     
     ! setup <c_mk| H |c_nk> for frozen manifold
     do ibnd = 1, QO_nbndk(ik)
        itruebnd = QO_bandindex(ibnd, ik)
        Enmk_temp(ibnd, ibnd) = cmplx(et(itruebnd, ik), 0.d0)

!!        !! debug: check if h_psi on evc_occupied_states gives the same eigenvalues from SCF calculation
!!        wfc_temp = cmplx(0.d0, 0.d0)
!!        call QO_h_psi (npwx, npw, 1, evc(1, ibnd), wfc_temp)
!!        temp_enk = cmplx( dble(ZDOTC (npwx*npol, evc(1, ibnd), 1, wfc_temp, 1)), 0.d0 )
!!#ifdef __PARA
!!        ! recover over G parallelization (intra_pool)
!!        call mp_sum (temp_enk, intra_pool_comm)
!!        write(stdout, *) 'et(ibnd) = ', ibnd, dble(Enmk_temp(ibnd, ibnd)), temp_enk
!!#endif
        
     end do
     QO_Enmk( :, :, ik) = Enmk_temp(:,:)
     
     write(stdout, '(8x, A)') 'QO_Enmk (diagonal part without re-orthogonalization):'
     do i = 1, natomwfc
        write(stdout, '(12x, 2i6, f10.6)') ik, i, dble(QO_Enmk(i, i, ik))
     end do
     write(stdout, '(A)') ''
     
     if (QO_flag_diagonalize_Hamiltonian) then
        n_cmk = natomwfc - QO_nbndk(ik)
        if (n_cmk>0) then
           allocate( Enmk_c( n_cmk, n_cmk ) )
           allocate( work_c( n_cmk, n_cmk ) )
           allocate( e_c( n_cmk) )
           QO_wfc(:,:) = cmplx(0.d0, 0.d0)
           Enmk_c(:,:) = cmplx(0.d0,0.d0)
           work_c(:,:) = cmplx(0.d0,0.d0)
           e_c(:) = 0.d0
           !
           do i = 1, n_cmk
              do j = 1, n_cmk
                 Enmk_c(i, j) = Enmk_temp(i+QO_nbndk(ik), j+QO_nbndk(ik))
              end do
           end do
           ! 
           ! calculate eigenvector and eigenvalues of Enmk_c matrix: [work, e] = eig(Enmk_c)
           !           
           call cdiagh (n_cmk, Enmk_c, n_cmk, e_c, work_c)
           ! write(stdout, *) dble(e_c)           
           !
           do i = 1, n_cmk
              do j = i, n_cmk
                 Enmk_temp(i+QO_nbndk(ik), j+QO_nbndk(ik)) = 0.d0
                 Enmk_temp(j+QO_nbndk(ik), i+QO_nbndk(ik)) = 0.d0
              end do
              Enmk_temp(i+QO_nbndk(ik), i+QO_nbndk(ik)) = e_c(i)
              do j = 1, n_cmk
                 QO_wfc(:, i) = QO_wfc(:, i) + QO_cmk(:, j+QO_nbndk(ik)) * work_c(j, i)
              end do
           end do
           QO_cmk(:, (QO_nbndk(ik)+1) : natomwfc) = QO_wfc(:, 1:n_cmk)
           QO_Enmk(:, :, ik) = Enmk_temp(:,:)
           !
           write(stdout, '(8x, A)') 'QO_Enmk (diagonal part after re-orthogonalization): '
           do i = 1, natomwfc
              write(stdout, '(12x, 2i6, f10.6)') ik, i, dble(QO_Enmk(i, i, ik))
           end do
           write(stdout, '(A)') ''
           !
           Enmk_temp(:, :) = cmplx(0.d0, 0.d0)
           Enmk_temp(:,:)  = QO_Omega( :, :, ik )
           QO_Omega( (QO_nbndk(ik)+1):natomwfc, :, ik ) = cmplx(0.d0, 0.d0)
           do i = 1, n_cmk
              do j = 1, n_cmk
                 QO_Omega( QO_nbndk(ik)+i, :, ik ) = QO_Omega( QO_nbndk(ik)+i, :, ik ) + &
                      Enmk_temp(QO_nbndk(ik)+j, :) * conjg(work_c(j, i))
              end do
           end do           
           !
           QO_wfc(:,:) = cmplx(0.d0, 0.d0)
           Enmk_temp(:,:) = cmplx(0.d0, 0.d0)
           !
           deallocate (Enmk_c, work_c, e_c)
        end if
     end if
     
!--------------------------------------------------------------------------------------------------------
!! the following lines are commented since we need to do Lowdin's transformation before outputing any QOs
!     ! form QO(:, j) at ik
!     do i = 1, natomwfc
!        QO_wfc(:,i) = cmplx(0.d0, 0.d0)
!        do ibnd = 1, natomwfc
!           QO_wfc(:, i) = QO_wfc(:, i) +  QO_cmk(:, ibnd) * QO_Omega(ibnd, i, ik)
!        end do
!     end do
!     
!     ! write QO wfc on unit iunsat
!     call davcio (QO_wfc, nwordatwfc, iunsat, ik, 1)  
!--------------------------------------------------------------------------------------------------------
!     call davcio (QO_cmk, nwordatwfc, iunat, ik, 1)
     
     !
     ! Lowdin's orthonormalization
     !
     if ( QO_flag_orthonormalized ) then
        ! 
        ! calculate O(i,j) = QO_Omega(ibnd, i)' * QO_Omega(ibnd, j)
        ! 
        overlap(1:natomwfc, 1:natomwfc) = cmplx(0.d0, 0.d0)     
        call zgemm('C', 'N', natomwfc, natomwfc, natomwfc, &
             (1.d0, 0.d0), QO_Omega(1,1,ik), natomwfc, QO_Omega(1,1,ik), natomwfc, &
             (0.d0, 0.d0), overlap(1,1), natomwfc )
        ! 
        ! calculate eigenvector and eigenvalues of O matrix: [work, e] = eig(O)
        ! 
        work(:,:) = cmplx(0.d0,0.d0)
        e(:) = 0.d0
        call cdiagh (natomwfc, overlap, natomwfc, e, work)
        ! 
        ! calculate (O^{-1/2})_ij = work(i,k) * 1/dsqrt(e(k)) * work(j,k)'
        ! 
        do i = 1, natomwfc
           if (e(i) < 1.d-10) then
              write(stdout, '(A, 2i8, e(i))') 'ik, ieigen, eigenvalue', ik, i, e(i)
!!              call errore ('QO', 'overlap matrix has linearly dependent eigenvectors', 1)
           end if
           e(i) = 1.d0 / dsqrt( e(i) ) 
        enddo
        do i = 1, natomwfc
           do j = i, natomwfc
              overlap (i, j) = (0.d0, 0.d0) 
              do k = 1, natomwfc 
                 overlap(i, j) = overlap(i, j) + e(k) * work(j, k) * conjg(work (i, k) ) 
              enddo
              if (j /= i) overlap (j, i) = conjg (overlap (i, j)) 
           enddo
        enddo
        ! 
        ! calculate QO_Omega_NEW(ibnd, j, ik) = (O^{-1/2})_(j,k) * QO_Omega_OLD(ibnd, k, ik)
        ! 
        overlap2 = 0.d0
        do ibnd = 1, natomwfc
           do j  = 1, natomwfc
              do k = 1, natomwfc
                 overlap2(ibnd, j) = overlap2(ibnd, j) + overlap(j, k) * QO_Omega(ibnd, k, ik)
              end do
           end do
        end do
        QO_Omega(1:natomwfc, 1:natomwfc, ik) = overlap2(1:natomwfc, 1:natomwfc)
        
     end if
     
     xkprime(1:3) = 0.d0
     do i = 1, 3
        do j = 1, 3
           xkprime(i) = xkprime(i) + xk(j, ik) * at(j, i)
        end do
     end do
              
     ! phase factor: exp( + i * k * r) :  NOT epr( - ikr) !
     phase = matmul(Xgrid, xk(1:3, ik)) * cmplx(0.d0, 1.d0) * tpi
     phase = exp(phase)
              
     ! form QO(:, j) at ik in both G space
     ! and transform into real space
     do i = 1, natomwfc        
        QO_wfc(:, i) = cmplx(0.d0, 0.d0)
        do ibnd = 1, natomwfc
           QO_wfc(:, i) = QO_wfc(:, i) +  &
                QO_cmk(:, ibnd) * QO_Omega(ibnd, i, ik)
        end do
        
        ! transform QO contracted array in G space into full array in G space
        do ip = 1, npol
           istart = npwx*(ip-1)+1
           iend   = npwx*ip
           psic(:) = cmplx(0.d0, 0.d0)
           psic(nls(igk(1:npw))) = QO_wfc(istart: (istart+npw-1), i)
           ! transform QO from full array in G space to R space
           ! call cft3s (psic, nr1s, nr2s, nr3s, nr1sx, nr2sx, nr3sx, 2)
           call invfft ('Wave', psic, dffts)
           !! write (stdout, *) 'ik, wk(ik) = ', ik, wk(ik)
           psic(1:nrxxs) = psic(1:nrxxs) * phase(1:nrxxs)
           istart = nrxxs*(ip-1)+1
           iend   = nrxxs*ip
           QO_wfc_r(istart:iend, i, current_spin) = QO_wfc_r(istart:iend, i, current_spin) + psic(1:nrxxs)
        end do
        
     end do
     
     ! write QO wfc(ik) on unit iunsat
     call davcio (QO_wfc, nwordatwfc, iunsat, ik, 1)  
     
  end do
  
  ! write QO_wfc_r on unit iunsatreal
  call davcio (QO_wfc_r, nwordatwfcr, iunsatr, 1, 1)
  
  if (allocated(wfcatom))      deallocate (wfcatom)
  if (allocated(wfcatom_para)) deallocate (wfcatom_para )
  if (allocated(wfcatom_perp)) deallocate (wfcatom_perp )
  if (allocated(work))         deallocate (work)
  if (allocated(overlap))      deallocate (overlap)
  if (allocated(overlap2))     deallocate (overlap2)
  if (allocated(e))            deallocate (e)
  if (allocated(W))            deallocate (W)
  if (allocated(Omega_old))    deallocate (Omega_old)
  if (allocated(proj))         deallocate (proj)
  if (allocated(projsum))      deallocate (projsum)
  if (allocated(wfc_temp))     deallocate (wfc_temp)
  if (allocated(wfc_temp2))    deallocate (wfc_temp2)
  if (allocated(Enmk_temp))    deallocate (Enmk_temp)
  if (allocated(AO_norm_k))    deallocate (AO_norm_k)
  if (allocated(psic))         deallocate (psic)
  if (allocated(phase))        deallocate (phase)
    
  !if ( gamma_only ) then
  !   if (allocated(rbecp)) deallocate (rbecp)
  !else
  !   if (noncolin) then
  !      if (allocated(becp_nc)) deallocate ( becp_nc ) 
  !   else
  !      if (allocated(becp)) deallocate ( becp ) 
  !   end if
  !end if
  
  if (ionode) then
     write(stdout, '(A)') '[QO] QO construction: end.'
     write(stdout, '(A)') '----------------------------------------------------------------------'
     write(stdout, '(A)') ''
  end if
  
  return
  
  call stop_clock ('construction')
  
end subroutine QO_construction








subroutine QO_tightbinding
  
  use kinds,      only : dp
  use io_global,  only : ionode, stdout
  use io_files,   only : iunat, iunsat, iunwfc, nwordatwfc, nwordwfc, iunigk, outdir, prefix
  use ions_base,  only : nat
  use basis,      only : natomwfc
  use klist,      only : nkstot, nks, xk, ngk, wk
  use ldaU,       only : swfcatom, U_projection
  use wvfct,      only : npwx, npw, igk, nbnd, et, current_k, g2kin
  use uspp,       only : nkb, vkb
  use becmod,     only : bec_type, allocate_bec_type, deallocate_bec_type, becp, calbec
  use noncollin_module,     only : noncolin, npol
  use lsda_mod,             only : nspin, lsda, isk, current_spin
  use wavefunctions_module, only : evc
  use global_version,       only : version_number
  use cell_base,  only : tpiba2, tpiba, at
  use constants,  only : bohr_radius_angs, au_gpa, tpi, amu_au
  use gvect,      only : g, ngm, ecutwfc
  use mp,         only : mp_barrier, mp_sum
  use mp_global,  only : intra_pool_comm, inter_pool_comm
  use control_flags, only : gamma_only
  
  implicit none
  
  complex(dp), external :: zdotc
  
  real(dp) :: ktmp, KRcell
  integer  :: current_cell, i, j, k , m1, m2, icell, jcell, kcell, ik
  complex(dp) :: tmp_phase
  integer  :: ounit, ispin
  character (len=256) :: fileout
  character (len=50)  :: mystring  
  character(len=256) :: pp_file
  real(dp),    allocatable :: QO_norm(:,:)
  real(dp) :: xkprime(3)
  complex(dp), allocatable :: zmatrix(:,:)
  real(dp) :: wk_tot
  
  
  call start_clock ('tightbinding')

  !! ============================================================================ !!  
  if (ionode) then
     write(stdout, '(A)') ''
     write(stdout, '(A)') '----------------------------------------------------------------------'
     write(stdout, '(A)') '[QO] tightbinding: start ...'
     write(stdout, '(A)') ''
  end if
  write(stdout, '(1x, A, 4i8)') 'cell: ',  ncellx, ncelly, ncellz, ncell
  write(stdout, '(1x, A, i8)') 'nspin_general: ', nspin_general
   
  QO_H = cmplx(0.d0, 0.d0)
  QO_S = cmplx(0.d0, 0.d0)
  
  allocate(QO_norm(natomwfc, nspin_general) )
  
  allocate(zmatrix(natomwfc, natomwfc))
  zmatrix = cmplx(0.d0, 0.d0)
  
  call weights()
  
  wk_tot = 0.d0
  do ik = 1, nks
     
     current_k = ik     
     current_spin = 1
     if (lsda) current_spin = isk(ik)
     if (noncolin) current_spin = 1
     write(stdout, '(1x, A, 2i6)') 'ik, ispin = ', ik, current_spin
     
     call gk_sort (xk (1, ik), ngm, g, ecutwfc / tpiba2, npw, igk, g2kin)
     g2kin(:) = g2kin(:) * tpiba2
     
     
     xkprime(1:3) = 0.d0
     do i = 1, 3
        do j = 1, 3
           xkprime(i) = xkprime(i) + xk(j, ik) * at(j, i)
        end do
     end do
     
     zmatrix = cmplx(0.d0, 0.d0)
     QO_OEO(1:natomwfc, 1:natomwfc, ik) = cmplx(0.d0, 0.d0)     
     call zgemm('N', 'N', natomwfc, natomwfc, natomwfc, &
          (1.d0, 0.d0), QO_Enmk(1,1,ik),  natomwfc, QO_Omega(1,1,ik), natomwfc, &
          (0.d0, 0.d0), zmatrix, natomwfc )
     call zgemm('C', 'N', natomwfc, natomwfc, natomwfc, &
          (1.d0, 0.d0), QO_Omega(1,1,ik), natomwfc, zmatrix(1,1), natomwfc, &
          (0.d0, 0.d0), QO_OEO(1,1,ik), natomwfc )
     
     QO_OO(1:natomwfc, 1:natomwfc, ik) = cmplx(0.d0, 0.d0)     
     call zgemm('C', 'N', natomwfc, natomwfc, natomwfc, &
          (1.d0, 0.d0), QO_Omega(1,1,ik), natomwfc, QO_Omega(1,1,ik), natomwfc, &
          (0.d0, 0.d0), QO_OO(1,1,ik), natomwfc )
     
     current_cell = 0
     do icell = -ncellhalfx, ncellhalfx
        do jcell = -ncellhalfy, ncellhalfy
           do kcell = -ncellhalfz, ncellhalfz
              current_cell = current_cell + 1
              
              KRcell =  dble(icell) * xkprime(1)  + dble(jcell) * xkprime(2)  + dble(kcell) * xkprime(3)
              tmp_phase = wk(ik) * exp( cmplx(0.d0, 1.d0) * tpi * KRcell)
              
              QO_S(current_cell, :, :, current_spin) = &
                   QO_S(current_cell, :, :, current_spin) + &
                   tmp_phase * QO_OO(:, :, ik)
              
              QO_H(current_cell, :, :, current_spin) = &
                   QO_H(current_cell, :, :, current_spin) + &
                   tmp_phase * QO_OEO(:, :, ik)
              
           end do
        end do
     end do
     
     wk_tot = wk_tot + wk(ik)

  end do
  
  ! QO_H = dble ( QO_H )
  ! QO_S = dble ( QO_S )

#ifdef __PARA  
  ! recover over kpt parallelization (inter_pool)
  call mp_sum (QO_H, inter_pool_comm)
  call mp_sum (QO_S, inter_pool_comm)
  call mp_sum (wk_tot, inter_pool_comm)
#endif
  
#ifdef __PARA
  call mp_barrier()
#endif

  QO_H = QO_H / wk_tot *ryd_to_ev
  QO_S = QO_S / wk_tot
      
!  if (nspin == 1) then
!     QO_H = QO_H / 2.d0 *ryd_to_ev
!     QO_S = QO_S / 2.d0
!  else
!     QO_H = QO_H *ryd_to_ev
!     QO_S = QO_S
!  endif
  
  
  if (QO_flag_normalized) then
     
     do ispin = 1, nspin_general
        do i = 1, natomwfc
           QO_norm(i, ispin) = dble(QO_S( nint(real(ncell+1)/2.d0), i, i, ispin))
        end do
     end do
     
     do ik = 1, nks
        current_k = ik        
        current_spin = 1
        if (lsda) current_spin = isk(ik)
        if (noncolin) current_spin = 1
        do i = 1, natomwfc
           QO_Omega(:, i, ik) = QO_Omega(:, i, ik) /  dsqrt( QO_norm(i, current_spin) )
        end do
     end do
        
     do ispin = 1, nspin_general
        do i = 1, natomwfc
           do j = 1, natomwfc
              QO_H(:, i, j, ispin) = QO_H(:, i, j, ispin) / &
                   dsqrt( QO_norm(i, ispin) ) / &
                   dsqrt( QO_norm(j, ispin) )
           end do
        end do
     end do
     
     ! QO_S has to be renormalized in the end 
     do ispin = 1, nspin_general
        do i = 1, natomwfc
           do j = 1, natomwfc
              QO_S(:, i, j, ispin) = QO_S(:, i, j, ispin) / &
                   dsqrt( QO_norm(i, ispin) ) / &
                   dsqrt( QO_norm(j, ispin) )
           end do
        end do
     end do
     

  end if
  
  if (ionode) then
     write(stdout, '(A)') ''
     write(stdout, '(A)') '[QO] tightbinding: end.'
     write(stdout, '(A)') '----------------------------------------------------------------------'
     write(stdout, '(A)') ''
  end if
  
#ifdef __PARA
  call mp_barrier()
#endif
  
  deallocate(QO_norm)
  deallocate(zmatrix)
  
  call stop_clock ('tightbinding')
  
  return
  
end subroutine QO_tightbinding


subroutine QO_write2netcdf()
!! ================================================================================================== !!
  use netcdf
  use typeSizes
!! ================================================================================================== !!
  
  use kinds,          only : dp 
  use pwcom
  use global_version, only : version_number
  use wavefunctions_module,  only : evc
  use io_files,       only : nd_nmbr, outdir, prefix, iunwfc, nwordwfc, pseudo_dir, psfile, find_free_unit
  use io_global,      only : ionode, stdout
  use cell_base,      only : omega, alat, at, bg, tpiba, tpiba2
  use ions_base,      only : atm, nat, ityp, tau, nsp, zv
  use mp_global,      only : nproc, nproc_pool, mpime, my_pool_id, intra_pool_comm, inter_pool_comm
  use mp,             only : mp_sum, mp_barrier
  use uspp_param,     only : upf
  use scf,            only : rho, v, vrs, vltot
  use fft_base,       only : grid_gather, cgather_smooth
  use atom,           only : msh, rgrid
  use symm_base,      only : invsym, sname, ftau, s, nsym
  use noncollin_module, only : noncolin, npol
  use grid_dimensions, only: nr1, nr2, nr3, nr1x, nr2x, nr3x, nrxx
  use smooth_grid_dimensions, only : nr1s, nr2s, nr3s, nr1sx, nr2sx, nr3sx, nrxxs
  
  implicit none
  
  CHARACTER(5), PARAMETER :: fmt_name="QEXPT"
  CHARACTER(5), PARAMETER :: fmt_version="1.1.0"

  character(len=256) :: pp_file
  
  integer  :: i, j, k, ig, ik, ibnd, na, ngg,ig_, ierr, nb, nt, ic, it
  integer, allocatable :: kisort(:)
  real(dp) :: xyz(3), tmp(3)
  integer  :: npool, nkbl, nkl, nkr, npwx_g
  integer  :: ike, iks, npw_g, ispin, local_pw
  
  real(dp) :: wfc_scal 
  logical  :: twf0, twfm
  
  real(dp), allocatable :: aux(:)
  
!! ================================================================================================== !!
! NETCDF file      
  integer :: nc_file_id
  character(len=200) :: nc_file 
  
!     NETCDF dimensions: length
  integer ::   nband, nbzk, nibzk, natom, nionstep, nplanewave, linvsym
  integer ::   natomwfcpol
  integer ::   nsoftdim1, nsoftdim2, nsoftdim3, nharddim1, nharddim2, nharddim3, nmaingrid
  integer, parameter :: dim1= 1, dim2= 2, dim3= 3, dim4= 4, dim5= 5, dim6= 6, dim7= 7
  integer, parameter :: nlengthstr = 200
!  integer ::   nrxx_total, nrxxs_total
  integer ::   nb_max, nmesh_max  
  
  !     NETCDF dimensions: ID
  integer ::   dim1_id, dim2_id, dim3_id, dim4_id, dim5_id, dim6_id, dim7_id, real_complex_id
  integer ::   nionstep_id, natom_id, nspin_id, nband_id, ibzk_current_id
  integer ::   nsoftdim1_id, nsoftdim2_id, nsoftdim3_id, nharddim1_id, nharddim2_id, nharddim3_id
  integer ::   nibzk_id, nbzk_id, nsp_id, nsym_id, nplanewave_id, nmaingrid_id
  integer ::   nlengthstr_id
  integer ::   dim1_tmpid, dim3_tmpid, dim4_tmpid, real_complex_tmpid, nplanewave_tmpid, nband_tmpid
  integer ::   dim3_tmpid2, dim4_tmpid2, nplanewave_tmpid2
  integer ::   natomwfc_id, ncell_id, ncellx_id, ncelly_id, ncellz_id
  integer ::   npol_id, nspin_general_id
  integer ::   natomwfcpol_id
  integer ::   nr1x_id,  nr2x_id,  nr3x_id,  nrxx_id,   nrxx_total_id, nrxxs_total_id
  integer ::   nr1sx_id, nr2sx_id, nr3sx_id, nrxxs_id
  integer ::   nchix_id
  integer ::   nmesh_max_id
  integer ::   ncell_plot_id, ncellx_plot_id, ncelly_plot_id, ncellz_plot_id
  
  !     NETCDF variables: ID
  integer ::   Kpoint_Mesh_ID, Kpoint_MeshShift_ID, Kpoint_SpinIndex_ID
  integer ::   Kpoint_IBZK_ID, Kpoint_IBZKWeight_ID, Kpoint_IBZKCurrent_ID, Kpoint_IBZK_Length_ID
  integer ::   Kpoint_IBZKCurrent_tmpID, ibzk_current_id2, Kpoint_IBZK_Length_tmpID
  integer ::   UnitCell_ID, LatticeParameter_ID
  integer ::   Atom_Type_ID, Atom_Position_ID, Atom_Force_ID
  integer ::   Element_ID, Element_Valence_ID, Element_Potential_ID, Element_AtomicNumber_ID, Element_NumberOfAtomicOrbitals_ID
  integer ::   Element_RadialWavefunction_LogMesh_ID, Element_RadialWavefunction_ID
  
  integer ::   Symmetry_Inversion_ID, Symmetry_Name_ID, Symmetry_Rotation_ID, Symmetry_Translation_ID
  integer ::   Hamiltonian_MainGrid_ID, Hamiltonian_MainGrid_GSquare_ID
  integer ::   Hamiltonian_EigenValue_ID
  integer ::   Hamiltonian_Wavefunction_ID, Hamiltonian_Wavefunction_Grid_ID
  integer ::   Hamiltonian_Wavefunction_SPsi_ID,Hamiltonian_Wavefunction_SPsi_tmpID
  integer ::   Hamiltonian_Wavefunction_Cutoff_ID, Hamiltonian_ChargeDensity_Cutoff_ID
  integer ::   Hamiltonian_Wavefunction_tmpID, Hamiltonian_Wavefunction_Grid_tmpID
  integer ::   Hamiltonian_Wavefunction_Prefix_ID, Hamiltonian_Wavefunction_Grid_Prefix_ID
  integer ::   Hamiltonian_FermiLevel_ID, Hamiltonian_OccupationNumber_ID, Hamiltonian_TotalElectron_ID
  integer ::   Hamiltonian_ChargeDensity_ID, Hamiltonian_EffectivePotential_ID
  integer ::   QO_PseudoatomicOrbital_nlm_ID, QO_PseudoatomicOrbital_rescalingfactor_ID
  integer ::   QO_H_ID, QO_S_ID, QO_E_window_max_ID, QO_nfrozenband_ID, QO_Omega_ID, QO_bandindex_ID
  integer ::   QO_OO_ID, QO_OEO_ID, QO_Enmk_ID
  integer ::   Bloch_projsum_ID, AO_norm_ID
  
  integer ::   nc_nonzerowave(nkstot)
  integer  ::  atomicnumber(nsp), elements_nao(nsp)
  real(DP) ::  trans(3,nsym)
  character(len=3)   :: elements(nsp)
  character(len=nlengthstr) :: pseudopotential_path
  character(len=nlengthstr) :: pseudopotential_name(nsp)
  character(len=nlengthstr) :: symmetry_name(200)
  
  integer,  external    :: atomic_number
  real(dp), allocatable :: f_occupation(:,:)
  real(dp), allocatable :: Omega_real_complex(:,:,:,:)
  real(dp), allocatable :: QO_H_real_complex(:,:,:,:,:)
  real(dp), allocatable :: QO_S_real_complex(:,:,:,:,:)
  real(dp), allocatable :: OO_real_complex(:,:,:,:)
  real(dp), allocatable :: OEO_real_complex(:,:,:,:)
  real(dp), allocatable :: QO_Enmk_real_complex(:,:,:,:)
  real(dp), allocatable :: r(:,:), chi(:, :, :)
  real(dp), allocatable :: rho_tot(:,:), vrs_tot(:,:)
  !! ================================================================================================== !!
            
  call start_clock ('write2netcdf')
  
  if (ionode) then
     write(stdout, '(A)') ''
     write(stdout, '(A)') '----------------------------------------------------------------------'
     write(stdout, '(A)') '[QO] write2netcdf: start ...'
     write(stdout, '(A)') ''
  end if
  
  ! find out the global number of G vectors: ngm_g  
  ngm_g = ngm
  ! recover over G parallelization (intra_pool)
  call mp_sum( ngm_g , intra_pool_comm )
  
  allocate(f_occupation(nbnd, nkstot))
  f_occupation(:,:) = 0.d0
  do ik=1, nks
     if ( wk(ik) == 0.d0 ) then
        f_occupation(1:nbnd, ik) = wg(1:nbnd,ik)        
     else
        f_occupation(1:nbnd, ik) = wg(1:nbnd,ik) / wk(ik)
     end if
  enddo
  if (nspin == 2) then
     f_occupation = f_occupation / 2.d0
  end if
  
  allocate(rho_tot(nrxx_total, nspin))
  allocate(vrs_tot(nrxx_total, nspin))

  natomwfcpol = natomwfc     
  
#ifdef __PARA  
  ! recover over kpt parallelization (inter_pool)
  allocate(aux (nrxx) )
  call mp_sum (rho%of_r, inter_pool_comm)
  do ispin = 1, nspin
     call grid_gather (rho%of_r(:,ispin),   rho_tot(:,ispin))
     
     aux(:) = vltot(:) + v%of_r(:, ispin)
     call mp_sum (aux, inter_pool_comm)
     call grid_gather (aux,                 vrs_tot(:,ispin))
  end do
  deallocate(aux)
  
  !   vectors et and proj are distributed across the pools 
  !   collect data for all k-points to the first pool
  ! 
  call poolrecover (et,            nbnd, nkstot, nks) 
  call poolrecover (f_occupation,  nbnd, nkstot, nks)
  call poolrecover (QO_Omega, 2*natomwfcpol*natomwfcpol, nkstot, nks)
  call poolrecover (QO_OO,    2*natomwfcpol*natomwfcpol, nkstot, nks)
  call poolrecover (QO_OEO,   2*natomwfcpol*natomwfcpol, nkstot, nks)
  call poolrecover (QO_Enmk,  2*natomwfcpol*natomwfcpol, nkstot, nks)
  call poolrecover (Bloch_projsum, nbnd, nkstot, nks)
  call ipoolrecover (QO_bandindex,  nbnd, nkstot, nks)
  call ipoolrecover (QO_nbndk,      1,    nkstot, nks)
#else
  rho_tot = rho%of_r
  do ispin = 1, nspin
     vrs_tot(:,ispin) = vltot(:) + v%of_r(:, ispin)
  end do
#endif      
  
#ifdef __PARA 
  call mp_barrier()
#endif
  
  if( ionode ) then
     !! ================================================================================================== !!
     
     nionstep = 1
     natom = nat
     nspin = nspin
     
     nband = nbnd
     nibzk = nkstot
     if ( dble(nk1*nk2*nk3) > 0.d0) then
        nbzk  = nk1*nk2*nk3
     else
        nbzk  = 1
     end if
     nsp   = nsp
     nsym  = nsym
     nplanewave = npwx   !! need to assign later
     nsoftdim1 = nr1s
     nsoftdim2 = nr2s
     nsoftdim3 = nr3s
     nharddim1 = nr1
     nharddim2 = nr2
     nharddim3 = nr3
     nharddim2 = nr2
     nharddim3 = nr3
     nmaingrid = ngm_g
     
     nmesh_max = maxval(msh(1:nsp))
     
     allocate(r(nmesh_max, nsp))
     allocate(chi(nmesh_max, nchix_max, nsp))
     r(:,:)     = 0.d0
     chi(:,:,:) = 0.d0
     do nt = 1, nsp
        r(1:msh(nt), nt) = rgrid(nt)%r(1:msh(nt))
        do nb = 1, upf(nt)%nwfc
           if (upf(nt)%oc(nb) >= 0.d0) then
              chi( 1:msh(nt), nb, nt ) = upf(nt)%chi( 1:msh(nt), nb)
           end if
        end do
     end do
     
     pseudopotential_path = trim(pseudo_dir)
     
     do i=1, nsp 
        elements(i)  = atm(i)
        pseudopotential_name(i) = trim(psfile(i))
        atomicnumber(i) = atomic_number(trim(atm(i)))
        elements_nao(i) = n_element_atomic_orbital(i)
     enddo
     
     if (invsym) then
        linvsym = 1
     else
        linvsym = 0
     endif
     
     do i=1, nsym
        symmetry_name(i) = trim(sname(i))
        tmp(1) = ftau(1,i) / dble( nr1 )
        tmp(2) = ftau(2,i) / dble( nr2 )
        tmp(3) = ftau(3,i) / dble( nr3 )
        trans(1:3, i) = tmp(1:3)
     enddo
     
     
     allocate(Omega_real_complex(natomwfcpol, natomwfcpol, nibzk, 2))
     Omega_real_complex = 0.d0
     Omega_real_complex(:,:,:,1) = dble(QO_Omega(:,:,:))
     Omega_real_complex(:,:,:,2) = aimag(QO_Omega(:,:,:))
     
     allocate (QO_H_real_complex( ncell, natomwfc, natomwfc, nspin_general, 2 ))
     QO_H_real_complex = 0.d0
     QO_H_real_complex(:,:,:,:,1) = dble(QO_H(:,:,:,:))
     QO_H_real_complex(:,:,:,:,2) = aimag(QO_H(:,:,:,:))
     
     allocate (QO_S_real_complex( ncell, natomwfc, natomwfc, nspin_general, 2 ))
     QO_S_real_complex = 0.d0
     QO_S_real_complex(:,:,:,:,1) = dble(QO_S(:,:,:,:))
     QO_S_real_complex(:,:,:,:,2) = aimag(QO_S(:,:,:,:))
     
     allocate(OO_real_complex(natomwfcpol, natomwfcpol, nibzk, 2))
     OO_real_complex = 0.d0
     OO_real_complex(:,:,:,1) = dble(QO_OO(:,:,:))
     OO_real_complex(:,:,:,2) = aimag(QO_OO(:,:,:))
     
     allocate(OEO_real_complex(natomwfcpol, natomwfcpol, nibzk, 2))
     OEO_real_complex = 0.d0
     OEO_real_complex(:,:,:,1) = dble(QO_OEO(:,:,:))
     OEO_real_complex(:,:,:,2) = aimag(QO_OEO(:,:,:))
     
     allocate(QO_Enmk_real_complex(natomwfcpol, natomwfcpol, nibzk, 2))
     QO_Enmk_real_complex = 0.d0
     QO_Enmk_real_complex(:,:,:,1) = dble(QO_Enmk(:,:,:))
     QO_Enmk_real_complex(:,:,:,2) = aimag(QO_Enmk(:,:,:))
     
     pp_file = trim(prefix)//".QO.nc"
     nc_file_id = find_free_unit()
     nc_file    = trim(outdir)//"/"//trim(pp_file)
     
     write(stdout, '(A)') " writing netcdf file to: "//trim(nc_file)
     
     !! CREATE MAIN NETCDF FILE
     call check(nf90_create(path = trim(nc_file), cmode = nf90_clobber, ncid = nc_file_id)) 
     
     !! DEFINE THE NETCDF FILE INFORMATION
     call check(nf90_put_att(nc_file_id, nf90_global, "author", &
          "Xiaofeng Qian (MIT, 2008)"))
     call check(nf90_put_att(nc_file_id, nf90_global, "title", &
          "NetCDF file of Quantum-ESPRESSO package   version: "//trim(version_number) ))
     
     !! DEFINE DIMENSIONS
     call check( nf90_def_dim(ncid = nc_file_id, name = "dim1",  len=dim1, dimid= dim1_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "dim2",  len=dim2, dimid= dim2_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "dim3",  len=dim3, dimid= dim3_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "dim4",  len=dim4, dimid= dim4_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "dim5",  len=dim5, dimid= dim5_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "dim6",  len=dim6, dimid= dim6_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "dim7",  len=dim7, dimid= dim7_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "string_length",            len=nlengthstr, dimid= nlengthstr_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "real_complex",             len=2,        dimid= real_complex_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "number_of_ionic_steps",    len=nionstep, dimid= nionstep_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "number_of_dynamic_atoms",  len=natom,    dimid= natom_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "number_of_spins",          len=nspin,    dimid= nspin_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "number_of_general_spins",  len=nspin_general,    dimid= nspin_general_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "number_of_bands",          len=nband,    dimid= nband_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "softgrid_dim1",  len=nsoftdim1, dimid= nsoftdim1_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "softgrid_dim2",  len=nsoftdim2, dimid= nsoftdim2_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "softgrid_dim3",  len=nsoftdim3, dimid= nsoftdim3_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "hardgrid_dim1",  len=nharddim1, dimid= nharddim1_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "hardgrid_dim2",  len=nharddim2, dimid= nharddim2_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "hardgrid_dim3",  len=nharddim3, dimid= nharddim3_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "number_of_IBZ_kpoints", len=nibzk,      dimid= nibzk_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "number_of_BZ_kpoints",  len=nbzk,       dimid= nbzk_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "number_of_species",     len=nsp,        dimid= nsp_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "number_of_rotmatrices", len=nsym,       dimid= nsym_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "number_of_planewaves",  len=nplanewave, dimid= nplanewave_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "number_of_maingrids",   len=nmaingrid,  dimid= nmaingrid_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "ncell",     len=ncell,     dimid= ncell_id)  )
     call check( nf90_def_dim(ncid = nc_file_id, name = "ncellx",    len=ncellx,    dimid= ncellx_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "ncelly",    len=ncelly,    dimid= ncelly_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "ncellz",    len=ncellz,    dimid= ncellz_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "ncell_plot",     len=ncell_plot,     dimid= ncell_plot_id)  )
     call check( nf90_def_dim(ncid = nc_file_id, name = "ncellx_plot",    len=ncellx_plot,    dimid= ncellx_plot_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "ncelly_plot",    len=ncelly_plot,    dimid= ncelly_plot_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "ncellz_plot",    len=ncellz_plot,    dimid= ncellz_plot_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "natomwfcpol_id", len=natomwfcpol,     dimid= natomwfcpol_id))
     call check( nf90_def_dim(ncid = nc_file_id, name = "natomwfc",   len=natomwfc,  dimid= natomwfc_id    ))
     call check( nf90_def_dim(ncid = nc_file_id, name = "npol",       len=npol,      dimid= npol_id        ))
     call check( nf90_def_dim(ncid = nc_file_id, name = "nrxx_total", len=nrxx_total,dimid= nrxx_total_id  ))
     call check( nf90_def_dim(ncid = nc_file_id, name = "nrxxs_total",len=nrxxs_total,dimid= nrxxs_total_id))
     call check( nf90_def_dim(ncid = nc_file_id, name = "nrxx",      len=nrxx,      dimid= nrxx_id)  )
     call check( nf90_def_dim(ncid = nc_file_id, name = "nr1x",      len=nr1x,      dimid= nr1x_id)  )
     call check( nf90_def_dim(ncid = nc_file_id, name = "nr2x",      len=nr2x,      dimid= nr2x_id)  )
     call check( nf90_def_dim(ncid = nc_file_id, name = "nr3x",      len=nr3x,      dimid= nr3x_id)  )
     call check( nf90_def_dim(ncid = nc_file_id, name = "nrxxs",     len=nrxxs,     dimid= nrxxs_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "nr1sx",     len=nr1sx,     dimid= nr1sx_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "nr2sx",     len=nr2sx,     dimid= nr2sx_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "nr3sx",     len=nr3sx,     dimid= nr3sx_id) )
     call check( nf90_def_dim(ncid = nc_file_id, name = "number_of_L_components",  len=nchix_max,      dimid= nchix_id)     )
     call check( nf90_def_dim(ncid = nc_file_id, name = "number_of_radial_meshs",  len=nmesh_max,  dimid= nmesh_max_id) )
     
     !! DEFINE VARIABLES 
     !     KPOINT
     call check( nf90_def_var( nc_file_id, "Kpoint_Mesh", nf90_int, &
          dimids= (/ dim3_id /), varID= Kpoint_Mesh_ID) )
     call check( nf90_put_att( nc_file_id,  Kpoint_Mesh_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "Kpoint_MeshShift", nf90_double, &
          dimids= (/ dim3_id /), varID= Kpoint_MeshShift_ID) )
     call check( nf90_put_att( nc_file_id,  Kpoint_MeshShift_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "Kpoint_IBZK", nf90_double, &
          dimids= (/ dim3, nibzk_id /), varID= Kpoint_IBZK_ID) )
     call check( nf90_put_att( nc_file_id,  Kpoint_IBZK_ID,  "units",  "none") )      
     
     call check( nf90_def_var( nc_file_id, "Kpoint_IBZKWeight", nf90_double, &
          dimids= (/ nibzk_id /), varID= Kpoint_IBZKWeight_ID) )
     call check( nf90_put_att( nc_file_id,  Kpoint_IBZKWeight_ID,  "units",  "none") )  
     
     call check( nf90_def_var( nc_file_id, "Kpoint_SpinIndex", nf90_int, &
          dimids= (/ nibzk_id /), varID= Kpoint_SpinIndex_ID) )
     call check( nf90_put_att( nc_file_id,  Kpoint_SpinIndex_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "Kpoint_IBZK_Length", nf90_int, &
          dimids= (/ nibzk_id /), varID= Kpoint_IBZK_Length_ID) )
     call check( nf90_put_att( nc_file_id,  Kpoint_IBZK_Length_ID,  "units",  "none") )
     
     
     !     SYMMETRY
     call check( nf90_def_var( nc_file_id, "Symmetry_Name", nf90_char, &
          dimids= (/ nlengthstr_id, nsym_id /), varID= Symmetry_Name_ID) )
     call check( nf90_put_att( nc_file_id,  Symmetry_Name_ID,  "units",  "none") )      
     
     call check( nf90_def_var( nc_file_id, "Symmetry_Inversion", nf90_int, &
          dimids= (/ dim1 /), varID= Symmetry_Inversion_ID) )
     call check( nf90_put_att( nc_file_id,  Symmetry_Inversion_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "Symmetry_Rotation", nf90_int, &
          dimids= (/ dim3_id, dim3_id, nsym_id /), varID= Symmetry_Rotation_ID) )
     call check( nf90_put_att( nc_file_id,  Symmetry_Rotation_ID,  "units",  "none") )      
     
     call check( nf90_def_var( nc_file_id, "Symmetry_Translation", nf90_double, &
          dimids= (/ dim3_id, nsym_id /), varID= Symmetry_Translation_ID) )
     call check( nf90_put_att( nc_file_id,  Symmetry_Translation_ID,  "units",  "none") )      
     
     
     !     UNITCELL
     call check( nf90_def_var( nc_file_id, "UnitCell", nf90_double, &
          dimids= (/ dim3_id, dim3_id /), varID= UnitCell_ID) )
     call check( nf90_put_att( nc_file_id,  UnitCell_ID,  "units",  "Angstrom") )
     call check( nf90_def_var( nc_file_id, "LatticeParameter", nf90_double, &
          dimids= (/ dim1_id /), varID= LatticeParameter_ID) )
     call check( nf90_put_att( nc_file_id,  LatticeParameter_ID,  "units",  "Angstrom") )
     
     
     !     ATOM
     call check( nf90_def_var( nc_file_id, "Atom_Type", nf90_int, &
          dimids= (/ natom_id /), varID= Atom_Type_ID) )
     call check( nf90_put_att( nc_file_id,  Atom_Type_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "Atom_Position", nf90_double, &
          dimids= (/ dim3_id, natom_id, nionstep_id /), varID= Atom_Position_ID) )      
     call check( nf90_put_att( nc_file_id,  Atom_Position_ID,  "units",  "in the a_0 unit ( or, H(1,1) )") )
      
     
     !     ELEMENT
     call check( nf90_def_var( nc_file_id, "Element", nf90_char, &
          dimids= (/ dim3_id, nsp_id /), varID= Element_ID) ) 
     call check( nf90_put_att( nc_file_id,  Element_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "Element_AtomicNumber", nf90_int, &
          dimids= (/ nsp_id /), varID= Element_AtomicNumber_ID) ) 
     call check( nf90_put_att( nc_file_id,  Element_AtomicNumber_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "Element_NumberOfAtomicOrbitals", nf90_int, &
          dimids= (/ nsp_id /), varID= Element_NumberOfAtomicOrbitals_ID) ) 
     call check( nf90_put_att( nc_file_id,  Element_NumberOfAtomicOrbitals_ID,  "units",  "none") )          
     
     call check( nf90_def_var( nc_file_id, "Element_Valence", nf90_double,  &
          dimids= (/ nsp_id /), varID= Element_Valence_ID) )
     call check( nf90_put_att( nc_file_id,  Element_Valence_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "Element_Potential", nf90_char, &
          dimids= (/ nlengthstr_id, nsp_id /), varID= Element_Potential_ID) ) 
     call check( nf90_put_att( nc_file_id,  Element_Potential_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "Element_RadialWavefunction", nf90_double, &
          dimids= (/ nmesh_max_id, nchix_id, nsp_id /), varID= Element_RadialWavefunction_ID) ) 
     call check( nf90_put_att( nc_file_id,  Element_RadialWavefunction_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "Element_RadialWavefunction_LogMesh", nf90_double, &
          dimids= (/ nmesh_max_id, nsp_id /),   varID= Element_RadialWavefunction_LogMesh_ID) ) 
     call check( nf90_put_att( nc_file_id,  Element_RadialWavefunction_LogMesh_ID,  "units",  "none") )
     
     
!     HAMILTONIAN
!      call check( nf90_def_var( nc_file_id, "Hamiltonian_MainGrid", nf90_int, &
!           dimids= (/ dim3_id, nmaingrid_id /), varID= Hamiltonian_MainGrid_ID) ) 
!      call check( nf90_put_att( nc_file_id,  Hamiltonian_MainGrid_ID,  "units",  "none") )      
      
!      call check( nf90_def_var( nc_file_id, "Hamiltonian_MainGrid_GSquare", nf90_double, &
!           dimids= (/ nmaingrid_id /), varID= Hamiltonian_MainGrid_GSquare_ID) ) 
!      call check( nf90_put_att( nc_file_id,  Hamiltonian_MainGrid_GSquare_ID,  "units",  "(2*pi/*a)^2") )
      
     call check( nf90_def_var( nc_file_id, "Hamiltonian_EigenValue", nf90_double, &
          dimids= (/ nband_id, nibzk_id /), varID= Hamiltonian_EigenValue_ID) ) 
     call check( nf90_put_att( nc_file_id,  Hamiltonian_EigenValue_ID,  "units",  "eV") )
     
     call check( nf90_def_var( nc_file_id, "Hamiltonian_Fermilevel", nf90_double, &
          dimids= (/ dim1_id /), varID= Hamiltonian_Fermilevel_ID) ) 
     call check( nf90_put_att( nc_file_id,  Hamiltonian_Fermilevel_ID,  "units",  "eV") )
     
     call check( nf90_def_var( nc_file_id, "Hamiltonian_OccupationNumber", nf90_double, &
          dimids= (/ nband_id, nibzk_id /), varID= Hamiltonian_OccupationNumber_ID) ) 
     call check( nf90_put_att( nc_file_id,  Hamiltonian_OccupationNumber_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "Hamiltonian_TotalElectron", nf90_double, &
          dimids= (/ dim1_id /), varID= Hamiltonian_TotalElectron_ID) ) 
     call check( nf90_put_att( nc_file_id,  Hamiltonian_TotalElectron_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "Hamiltonian_Wavefunction_Cutoff", nf90_double, &
          dimids= (/ dim1_id /), varID= Hamiltonian_Wavefunction_Cutoff_ID) ) 
     call check( nf90_put_att( nc_file_id,  Hamiltonian_Wavefunction_Cutoff_ID,  "units",  "eV") )
      
     call check( nf90_def_var( nc_file_id, "Hamiltonian_ChargeDensity_Cutoff", nf90_double, &
          dimids= (/ dim1_id /), varID= Hamiltonian_ChargeDensity_Cutoff_ID) ) 
     call check( nf90_put_att( nc_file_id,  Hamiltonian_ChargeDensity_Cutoff_ID,  "units",  "eV") )
     
     call check( nf90_def_var( nc_file_id, "Hamiltonian_ChargeDensity", nf90_double, &
          dimids= (/ nrxx_total_id, nspin_id /), varID= Hamiltonian_ChargeDensity_ID) ) 
     call check( nf90_put_att( nc_file_id,  Hamiltonian_ChargeDensity_ID,  &
          "units",  "n_electron/A^3: it satisfies sum(rho) * volume / (N1*N2*N3) = n_electron") )
     
     call check( nf90_def_var( nc_file_id, "Hamiltonian_EffectivePotential", nf90_double, &
          dimids= (/ nrxx_total_id, nspin_id /), varID= Hamiltonian_EffectivePotential_ID) ) 
     call check( nf90_put_att( nc_file_id, Hamiltonian_EffectivePotential_ID ,  "units",  "eV") )
     
     !     QO
     call check( nf90_def_var( nc_file_id, "QO_E_window_max", nf90_double, &
          dimids= (/ dim1_id /), varID= QO_E_window_max_ID) ) 
     call check( nf90_put_att( nc_file_id,  QO_E_window_max_ID,  "units",  "eV") )
     
     call check( nf90_def_var( nc_file_id, "QO_H", nf90_double, &
          dimids= (/ ncell_id, natomwfcpol_id, natomwfcpol_id, nspin_general_id, real_complex_id /), varID= QO_H_ID) ) 
     call check( nf90_put_att( nc_file_id,  QO_H_ID,  "units",  "eV") )
     
     call check( nf90_def_var( nc_file_id, "QO_S", nf90_double, &
          dimids= (/ ncell_id, natomwfcpol_id, natomwfcpol_id, nspin_general_id, real_complex_id /), varID= QO_S_ID) )
     call check( nf90_put_att( nc_file_id,  QO_S_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "QO_PseudoatomicOrbital_nlm", nf90_int, &
          dimids= (/ nsp_id, nchix_id /), varID= QO_PseudoatomicOrbital_nlm_ID))
     call check( nf90_put_att( nc_file_id,  QO_PseudoatomicOrbital_nlm_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "QO_PseudoatomicOrbital_rescalingfactor", nf90_double, &
          dimids= (/ nsp_id, nchix_id /), varID= QO_PseudoatomicOrbital_rescalingfactor_ID))
     call check( nf90_put_att( nc_file_id,  QO_PseudoatomicOrbital_rescalingfactor_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "QO_Omega", nf90_double, &
          dimids= (/ natomwfcpol_id, natomwfcpol_id, nibzk_id, real_complex_id /), varID= QO_Omega_ID))
     call check( nf90_put_att( nc_file_id,  QO_Omega_ID,  "units",  "<psi_nk|S|QO>") )
     
     call check( nf90_def_var( nc_file_id, "QO_Sk", nf90_double, &
          dimids= (/ natomwfcpol_id, natomwfcpol_id, nibzk_id, real_complex_id /), varID= QO_OO_ID))
     call check( nf90_put_att( nc_file_id,  QO_OO_ID,   "units",  "<QO_k(Ii)|S|QO_k(Jj)>") )
     
     call check( nf90_def_var( nc_file_id, "QO_Hk", nf90_double, &
          dimids= (/ natomwfcpol_id, natomwfcpol_id, nibzk_id, real_complex_id /), varID= QO_OEO_ID))
     call check( nf90_put_att( nc_file_id,  QO_OEO_ID,  "units",  "<QO_k(Ii)|H|QO_k(Jj)>") )
     
     call check( nf90_def_var( nc_file_id, "QO_bandindex", nf90_int, &
          dimids= (/ nband_id, nibzk_id /), varID= QO_bandindex_ID) ) 
     call check( nf90_put_att( nc_file_id,  QO_bandindex_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "QO_nfrozenband", nf90_int, &
          dimids= (/ nibzk_id /), varID= QO_nfrozenband_ID))
     call check( nf90_put_att( nc_file_id,  QO_nfrozenband_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "Bloch_projsum", nf90_double, &
          dimids= (/ nband_id, nibzk_id /), varID= Bloch_projsum_ID) ) 
     call check( nf90_put_att( nc_file_id,  Bloch_projsum_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "AO_norm", nf90_double, &
          dimids= (/ natomwfc_id /), varID= AO_norm_ID) ) 
     call check( nf90_put_att( nc_file_id,  AO_norm_ID,  "units",  "none") )
     
     call check( nf90_def_var( nc_file_id, "QO_Enmk", nf90_double, &
          dimids= (/ natomwfcpol_id, natomwfcpol_id, nibzk_id, real_complex_id /), varID= QO_Enmk_ID))
     call check( nf90_put_att( nc_file_id,  QO_Enmk_ID,  "units",  "<C_mk| H |C_nk>") )     
     
     !! LEAVE DEFINITION MODE
     call check(nf90_enddef(nc_file_id))
     
     
     !! WRITE VARIABLES
     call check( nf90_put_var( nc_file_id,  Kpoint_Mesh_ID,          (/ nk1, nk2, nk3 /) ) )
     call check( nf90_put_var( nc_file_id,  Kpoint_MeshShift_ID,     (/ k1,  k2,  k3 /)  ) )
     call check( nf90_put_var( nc_file_id,  Kpoint_IBZK_ID,          xk(1:dim3,1:nkstot) ) )
     call check( nf90_put_var( nc_file_id,  Kpoint_IBZKWeight_ID,    wk(1:nkstot)        ) ) 
     call check( nf90_put_var( nc_file_id,  Kpoint_SpinIndex_ID,     isk(1:nkstot)       ) ) 
     
     call check( nf90_put_var( nc_file_id,  Symmetry_Name_ID,        symmetry_name(1:nsym)    ) )
     call check( nf90_put_var( nc_file_id,  Symmetry_Inversion_ID,   linvsym                  ) ) 
     call check( nf90_put_var( nc_file_id,  Symmetry_Rotation_ID,    s(1:dim3,1:dim3, 1:nsym) ) ) 
     call check( nf90_put_var( nc_file_id,  Symmetry_Translation_ID, trans(1:dim3, 1:nsym)    ) ) 
     
     call check( nf90_put_var( nc_file_id,  UnitCell_ID,             at(1:dim3,1:dim3) * alat * bohr_to_angstrom ) )
     call check( nf90_put_var( nc_file_id,  LatticeParameter_ID,     alat * bohr_to_angstrom ) )
     
     call check( nf90_put_var( nc_file_id,  Atom_Type_ID,            ityp(1:natom)          ) )
     call check( nf90_put_var( nc_file_id,  Atom_Position_ID,        tau(1:dim3,1:natom) * alat * bohr_to_angstrom  ) )
     
     !     WRITE ELEMENT
     call check( nf90_put_var( nc_file_id,  Element_ID,              elements(1:nsp)     ) )
     call check( nf90_put_var( nc_file_id,  Element_AtomicNumber_ID, atomicnumber(1:nsp) ) )
     call check( nf90_put_var( nc_file_id,  Element_NumberOfAtomicOrbitals_ID,      elements_nao(1:nsp) ) )
     call check( nf90_put_var( nc_file_id,  Element_Valence_ID,      zv(1:nsp)           ) )
     call check( nf90_put_var( nc_file_id,  Element_Potential_ID,    pseudopotential_name(1:nsp) ) )
     call check( nf90_put_var( nc_file_id,  Element_RadialWavefunction_LogMesh_ID,  r(1:nmesh_max, 1:nsp) * bohr_to_angstrom ) )
     call check( nf90_put_var( nc_file_id,  Element_RadialWavefunction_ID,          chi(1:nmesh_max, 1:nchix_max, 1:nsp)))
     
     !     WRITE HAMILTONAIN
     call check( nf90_put_var( nc_file_id,  Hamiltonian_EigenValue_ID,           et(1:nband, 1:nkstot) * ryd_to_ev     ) )
     call check( nf90_put_var( nc_file_id,  Hamiltonian_FermiLevel_ID,           ef * ryd_to_ev      ) )
     call check( nf90_put_var( nc_file_id,  Hamiltonian_OccupationNumber_ID,     f_occupation(1:nband, 1:nkstot) ) )
     call check( nf90_put_var( nc_file_id,  Hamiltonian_TotalElectron_ID,        nelec   ) )
     call check( nf90_put_var( nc_file_id,  Hamiltonian_Wavefunction_Cutoff_ID,  ecutwfc * ryd_to_ev) )
     call check( nf90_put_var( nc_file_id,  Hamiltonian_ChargeDensity_Cutoff_ID, dual*ecutwfc * ryd_to_ev) )
     call check( nf90_put_var( nc_file_id,  Hamiltonian_ChargeDensity_ID,        rho_tot(1:nrxx_total, 1:nspin ) / (bohr_to_angstrom**3)  ) )
     call check( nf90_put_var( nc_file_id,  Hamiltonian_EffectivePotential_ID,   vrs_tot(1:nrxx_total, 1:nspin )* ryd_to_ev / (bohr_to_angstrom**3) ) )
     
     !     WRITE QO
     call check( nf90_put_var( nc_file_id,  QO_E_window_max_ID, QO_E_window_max))
     call check( nf90_put_var( nc_file_id,  QO_H_ID,            QO_H_real_complex(1:ncell, 1:natomwfcpol, 1:natomwfcpol, 1:nspin_general, 1:2) ) )
     call check( nf90_put_var( nc_file_id,  QO_S_ID,            QO_S_real_complex(1:ncell, 1:natomwfcpol, 1:natomwfcpol, 1:nspin_general, 1:2) ) )
     call check( nf90_put_var( nc_file_id,  QO_PseudoatomicOrbital_nlm_ID, QO_PseudoatomicOrbital_nlm(1:nsp, 1:nchix_max) ) )
     call check( nf90_put_var( nc_file_id,  QO_PseudoatomicOrbital_rescalingfactor_ID, &
          QO_PseudoatomicOrbital_rescalingfactor(1:nsp, 1:nchix_max) ))
     call check( nf90_put_var( nc_file_id,  QO_Omega_ID,        Omega_real_complex(1:natomwfcpol, 1:natomwfcpol,1:nibzk,1:2)))
     call check( nf90_put_var( nc_file_id,  QO_OO_ID,           OO_real_complex(1:natomwfcpol, 1:natomwfcpol,1:nibzk,1:2)))
     call check( nf90_put_var( nc_file_id,  QO_OEO_ID,          OEO_real_complex(1:natomwfcpol, 1:natomwfcpol,1:nibzk,1:2)))
     call check( nf90_put_var( nc_file_id,  QO_bandindex_ID,    QO_bandindex(1:nbnd, 1:nibzk)))
     call check( nf90_put_var( nc_file_id,  QO_nfrozenband_ID,  QO_nbndk(1:nibzk)))
     call check( nf90_put_var( nc_file_id,  Bloch_projsum_ID,   Bloch_projsum(1:nbnd, 1:nibzk)))
     call check( nf90_put_var( nc_file_id,  AO_norm_ID,         AO_norm(1:natomwfc)))
     call check( nf90_put_var( nc_file_id,  QO_Enmk_ID,         QO_Enmk_real_complex(1:natomwfcpol, 1:natomwfcpol,1:nibzk,1:2) * ryd_to_ev ))
     
     call check( nf90_close( nc_file_id ))
     
     deallocate(Omega_real_complex)
     deallocate(QO_H_real_complex)
     deallocate(QO_S_real_complex)
     deallocate(OO_real_complex)
     deallocate(OEO_real_complex)
     deallocate(QO_Enmk_real_complex)
     deallocate(r)
     deallocate(chi)
     
     !! ================================================================================================== !!
  end if
  
#ifdef __PARA
  call mp_barrier()
#endif

  deallocate(rho_tot)
  deallocate(vrs_tot)
  deallocate(f_occupation)

  if (ionode) then
     write(stdout, '(A)') ''
     write(stdout, '(A)') '[QO] write2netcdf: end.'
     write(stdout, '(A)') '----------------------------------------------------------------------'
     write(stdout, '(A)') ''
  end if

  call stop_clock ('write2netcdf')
  
end subroutine QO_write2netcdf





subroutine check(status)
  use netcdf
  integer, intent ( in) :: status
  
  if(status /= nf90_noerr) then 
     write (*,'(A)') trim(nf90_strerror(status))
     call errore ('QO', 'netcdf error: status=', status)
  end if
end subroutine check







subroutine QO_output_orbital_kpoints
  
  use kinds,     only : DP
  use constants, only : tpi, bohr_radius_angs, au_gpa, tpi, amu_au
  use io_global, only : stdout, ionode
  use cell_base, only : omega, alat, at, bg, tpiba, tpiba2
  use ions_base, only : nat, atm, zv, tau, ntyp => nsp, ityp
  use gvect,     only : g, ngm, ecutwfc
  use smooth_grid_dimensions, only : nr1s, nr2s, nr3s, nr1sx, nr2sx, nr3sx, nrxxs
  use wvfct,     only : nbnd, npwx, npw, igk, wg, g2kin, current_k  
  use lsda_mod,  only : lsda, isk, nspin
  use klist,     only : xk, nks, nkstot, wk, ngk
  use basis,     only : natomwfc
  use io_files,  only : tmp_dir, prefix, iunat, iunsat, nwordatwfc, find_free_unit
  use mp_global, only : intra_pool_comm, inter_pool_comm
  use mp,        only : mp_sum
  use fft_base,  only : grid_gather, cgather_smooth
  use uspp_param,only : upf 
  use noncollin_module, only : noncolin, npol
  use control_flags,    only : noinv
  use gsmooth,   only : nls, nlsm
  USE fft_base,       ONLY : dffts
  USE fft_interfaces, ONLY : invfft, fwfft
  
  
  implicit none
  
  integer,  external  :: atomic_number
  integer             :: ounit , iorb, niorb, index_atom, it
  real(dp)            :: x0(3), ee1(3), ee2(3), ee3(3)
  complex(dp), allocatable  :: psic(:), wfcatom(:,:), psic_BV(:,:,:), phase(:), psic_total(:)
  character (len=256) :: fileout
  character (len=50)  :: mystring, mystring_spin, mystring_ipol, mystring_ik
  integer             :: current_spin, ip, istart, iend
  real(dp)     :: ktmp, KRcell, H(3,3), HH(3,3)
  real(dp)     :: xkprime(3)
  integer      :: current_cell, i, j, k , m1, m2, icell, jcell, kcell, ik_index, ik, ispin, icount, ibnd, ncell3_plot(3)
  complex(dp)  :: tmp_phase
  real(dp)     :: sumrho
  integer      :: n, iatom
  real(dp)     :: r_shift(3), atom_position(3, nat*ncell_plot)
  integer      :: atom_type(nat*ncell_plot)
  integer      :: atom_atomicnumber(nat*ncell_plot)
  character(3) :: atom_symbol(nat*ncell_plot)
  integer      :: natom, data_ispin, data_iorbital, data_iatom
  
  call start_clock ('output')
  
  ee1 = 0.d0
  ee2 = 0.d0
  ee3 = 0.d0
  
  if (ionode) then     
     write(stdout, '(A)') '----------------------------------------------------------------------'
     write(stdout, '(A)') '[QO] QO output: start ...'
     write(stdout, '(A)') ''
     ounit  = find_free_unit()
  endif
  
!  nrxxs_total = nr1sx * nr2sx * nr3sx
  allocate( psic_total (nrxxs_total))
  allocate( psic (nrxxs) )
  allocate( psic_BV( nr1sx*ncellx_plot, nr2sx*ncelly_plot, nr3sx*ncellz_plot ) )
  allocate( phase (nrxxs_total) )
  
  ! H for unitcell
  H(:,:) = 0.d0
  H(1:3, 1:3) = transpose(at(1:3,1:3))
  
  ! HH for supercell in Angstrom (plotting)
  ncell3_plot = (/ ncellx_plot, ncelly_plot, ncellz_plot /)
  HH(:,:) = 0.d0
  do i=1,3
     do j=1,3
        HH(j,i) = at(j,i) * dble(ncell3_plot(i)) *alat* bohr_radius_angs
     enddo
  enddo

  !! prepare the atom positions and atom type for netcdf output
  iatom = 1
  natom = ncell_plot * nat
  do i = 1, ncellx_plot
     do j = 1, ncelly_plot
        do k = 1, ncellz_plot
           
           r_shift = at(1:3, 1) * (i-1) * alat*bohr_radius_angs + &
                at(1:3, 2) * (j-1) * alat*bohr_radius_angs + &
                at(1:3, 3) * (k-1) * alat*bohr_radius_angs

           do n=1,nat

              atom_symbol(iatom) = atm(ityp(n))
              atom_type(iatom)   = ityp(n)
              atom_atomicnumber(iatom) = atomic_number(trim(atm(ityp(n))))
              ! positions are in Angstroms
              atom_position(1, iatom) = tau(1,n)*alat*bohr_radius_angs + r_shift(1)
              atom_position(2, iatom) = tau(2,n)*alat*bohr_radius_angs + r_shift(2)
              atom_position(3, iatom) = tau(3,n)*alat*bohr_radius_angs + r_shift(3)
              iatom = iatom + 1
              
           enddo
        end do
     end do
  end do
  
  
  x0     = 0.d0
  ee1(1) = 1.d0
  ee2(2) = 1.d0
  ee3(3) = 1.d0
  !
  ! Output QUASI-ATOMIC ORBITALS (QO)
  !
  ! QO_n(r) = \sum_{k} \sum_{L} u_{nk}(r + L) * exp(i*k*(r+L))
  !         = \sum_{k} u_{nk}(r) * exp(i*k*r) * \sum_{L} * exp(i*k*L)
  !
  do ispin = 1, nspin_general
     
     index_atom = 1
     if (noncolin) then
        niorb = QO_PseudoatomicOrbital_number(index_atom) * 2
     else
        niorb = QO_PseudoatomicOrbital_number(index_atom)
     end if
     
     do iorb = 1, natomwfc
        if (iorb > niorb) then
           index_atom = index_atom + 1
           if (noncolin) then
              niorb = niorb + QO_PseudoatomicOrbital_number(index_atom)*2
           else
              niorb = niorb + QO_PseudoatomicOrbital_number(index_atom)
           end if
        end if
        
        ! Plot out QOs of selected atoms in netcdf format
        if (QO_plot_atomflag(index_atom) > 0) then
           
        do ip = 1, npol
           
           istart = npwx*(ip-1)+1
           iend = npwx*ip
                 
           ! ----------------------------------------------------------------------------- !
           do ik_index = 1, QO_plot_number_of_kpoints
              ik = QO_plot_kpoints_index(ik_index)
              
              psic_BV = cmplx(0.d0, 0.d0)
        
              current_k = ik
              current_spin = isk(ik)
              if (noncolin) current_spin = 1
              
              if (current_spin == ispin) then
                 
                 ! calculate G vectors at k-point ik
                 call gk_sort (xk (1, ik), ngm, g, ecutwfc / tpiba2, npw, igk, g2kin)
                 g2kin(:) = g2kin(:) * tpiba2
                 
                 npw = ngk (ik)
                 xkprime(1:3) = 0.d0
                 do i = 1, 3
                    do j = 1, 3
                       xkprime(i) = xkprime(i) + xk(j, ik) * at(j, i)
                    end do
                 end do
                 
                 ! phase factor: exp( + i * k * r) :  NOT epr( - ikr) !
                 phase = matmul(Xgrid_total, xk(1:3, ik)) * cmplx(0.d0, 1.d0) * tpi
                 phase = exp(phase)
                 
                 ! read QO wfc (in G-vector representation) on unit iunsat
                 call davcio (QO_wfc, nwordatwfc, iunsat, ik, -1)  
                 
                 ! transform QO contracted array in G space into full array in G space
                 psic_total(:) = 0.d0              
                 psic(:) = cmplx(0.d0, 0.d0)
                 psic(nls(igk(1:npw))) = QO_wfc( istart: (istart+npw-1), iorb)
                 
                 ! transform QO from full array in G space to R space
                 ! call cft3s (psic, nr1s, nr2s, nr3s, nr1sx, nr2sx, nr3sx, 2)
                 call invfft ('Wave', psic, dffts)
#ifdef __PARA
                 call cgather_smooth ( psic, psic_total)
#endif
                 ! write(stdout, *) 'ik, wk =  ', ik, wk(ik)
                 ! psic_total(1:nrxxs_total) = psic_total(1:nrxxs_total) * wk(ik) * phase(1:nrxxs_total)
                 psic_total(1:nrxxs_total) = psic_total(1:nrxxs_total) * phase(1:nrxxs_total)
                 
                 do icell = - ncellhalfx_plot, ncellhalfx_plot
                    do jcell = - ncellhalfy_plot, ncellhalfy_plot
                       do kcell = - ncellhalfz_plot, ncellhalfz_plot
                          
                          KRcell =  &
                               dble(icell) * xkprime(1)  + &
                               dble(jcell) * xkprime(2)  + &
                               dble(kcell) * xkprime(3)
                          
                          tmp_phase = exp( cmplx(0.d0, 1.d0) * tpi * KRcell)
                          
                          psic_BV( &
                               ((icell+ncellhalfx_plot)*nr1sx+1) : (icell + ncellhalfx_plot +1)*nr1sx, &
                               ((jcell+ncellhalfy_plot)*nr2sx+1) : (jcell + ncellhalfy_plot +1)*nr2sx, &
                               ((kcell+ncellhalfz_plot)*nr3sx+1) : (kcell + ncellhalfz_plot +1)*nr3sx ) &
                               = psic_BV( &
                               ((icell+ncellhalfx_plot)*nr1sx+1) : (icell + ncellhalfx_plot +1)*nr1sx, &
                               ((jcell+ncellhalfy_plot)*nr2sx+1) : (jcell + ncellhalfy_plot +1)*nr2sx, &
                               ((kcell+ncellhalfz_plot)*nr3sx+1) : (kcell + ncellhalfz_plot+1)*nr3sx ) &
                               + tmp_phase * reshape( psic_total, (/nr1sx, nr2sx, nr3sx/) )
                          
                       end do ! do kcell = - ncellhalfz_plot, ncellhalfz_plot
                    end do ! do jcell = - ncellhalfy_plot, ncellhalfy_plot
                 end do !  do icell = - ncellhalfx_plot, ncellhalfx_plot

!! sum over different kpoints                 
!! #ifdef __PARA
!!            call mp_sum ( psic_BV, inter_pool_comm )
!! #endif

           if (.not. noinv) then
              psic_BV = dble(psic_BV)
           end if
           if (noncolin) then
              psic_BV = psic_BV / sqrt(dble(nr1sx*nr2sx*nr3sx))
           else
              if (nspin == 2) then
                 psic_BV = psic_BV / sqrt(dble(nr1sx*nr2sx*nr3sx))
              else  ! (nspin == 1)
                 psic_BV = psic_BV / sqrt(dble(nr1sx*nr2sx*nr3sx)) / dble(2.d0)
              end if
           end if
           
           ! ----------------------------------------------------------------------------- !
           if (ionode) then
              
              write(mystring, '(I6)') iorb
              write(mystring_spin, '(I6)') ispin
              write(mystring_ipol, '(I6)') ip
              write(mystring_ik,   '(I6)') ik
              
              if ( (trim(QO_flag_output_orbital_format) == 'XSF') .OR. &
                   (trim(QO_flag_output_orbital_format) == 'BOTH') ) then
                 
                 !! create XSF file
                 ! xsf file name 
                 fileout = trim(tmp_dir)//trim(prefix)//'.QO_'//trim(adjustl(mystring))//'_spin_'//trim(adjustl(mystring_spin))//'_ik_'//trim(adjustl(mystring_ik))//'.xsf'
                 if (npol > 1) then
                    fileout = trim(tmp_dir)//trim(prefix)//'.QO_'//trim(adjustl(mystring))//'_spin_'//trim(adjustl(mystring_spin))//'_ipol_'//trim(adjustl(mystring_ipol))//'_ik_'//trim(adjustl(mystring_ik))//'.xsf'
                 end if
                 write( stdout, '(A)') '  writing QO file to: '//trim(fileout)
                 
                 ! create xsf file
                 open (unit=ounit, file=fileout, form='formatted', status='unknown')
                 
                 ! output xsf structure
                 call xsf_struct_supercell (alat, at, nat, tau, atm, ityp, ounit, (/ncellx_plot, ncelly_plot, ncellz_plot/) )
                 
                 ! output xsf data_3d
                 call xsf_fast_datagrid_3d_supercell &
                      (dble(psic_BV), nr1s*ncellx_plot, nr2s*ncelly_plot, nr3s*ncellz_plot, &
                      nr1sx*ncellx_plot, nr2sx*ncelly_plot, nr3sx*ncellz_plot, &
                      at, alat, ounit, (/ncellx_plot, ncelly_plot, ncellz_plot/) )  
                 
                 ! close xsf file
                 close (unit=ounit)
              end if
                            
              if ( (trim(QO_flag_output_orbital_format) == 'NETCDF') .OR. &
                   (trim(QO_flag_output_orbital_format) == 'BOTH') ) then
                 data_iorbital  = iorb
                 data_ispin = ispin
                 data_iatom = &
                      ncellx_plot*ncelly_plot*ncellhalfz_plot * nat + &
                      ncellx_plot*ncellhalfy_plot * nat + &
                      ncellhalfx_plot * nat + index_atom
                 if (noncolin) data_ispin = 1
                 
                 !! create NetCDF file
                 ! netcdf file name
                 fileout = trim(tmp_dir)//trim(prefix)//'.QO_'//trim(adjustl(mystring))//'_spin_'//trim(adjustl(mystring_spin))//'_ik_'//trim(adjustl(mystring_ik))//'.nc'
                 if (npol >1) then
                    fileout = trim(tmp_dir)//trim(prefix)//'.QO_'//trim(adjustl(mystring))//'_spin_'//trim(adjustl(mystring_spin))//'_ipol_'//trim(adjustl(mystring_ipol))//'_ik_'//trim(adjustl(mystring_ik))//'.nc'
                 end if
                 write( stdout, '(A)') '  writing QO file to: '//trim(fileout)
                 !! write( stdout, *) ' (debug) data_iatom = ', data_iatom
                 call QO_output_netcdf_3d (fileout, psic_BV, &
                      nr1s*ncellx_plot, nr2s*ncelly_plot, nr3s*ncellz_plot, &
                      HH, natom, atom_symbol, atom_type, atom_atomicnumber, atom_position, &
                      data_iorbital, data_ispin, data_iatom, QO_flag_output_orbital_rcutoff )
              end if
              
              sumrho = sum(conjg(psic_BV) * psic_BV)
              write(stdout, '(A, 2i6, 2x, f24.12)') 'ispin, iorb, sumrho : ', ispin, iorb, sumrho
           end if   ! if (ionode)
           
           
           end if    ! if (current_spin == ispin) then
        end do    ! do ik = 1, nks
           
        end do !  do ip = 1, npol
        end if !! if (QO_plot_atomflag(index_atom) > 0)
        
     end do !  do iorb = 1, natomwfc
  end do ! do ispin = 1, nspin
  
  deallocate(psic_BV)
  deallocate(psic)
  deallocate(psic_total)
  deallocate(phase)
  
  if (ionode) then
     write(stdout, '(A)') ''     
     write(stdout, '(A)') '[QO] QO output: end.'
     write(stdout, '(A)') '----------------------------------------------------------------------'
     write(stdout, '(A)') ''     
  end if

  call stop_clock ('output')
  
end subroutine QO_output_orbital_kpoints



subroutine QO_output_orbital
  
  use kinds,     only : DP
  use constants, only : tpi, bohr_radius_angs, au_gpa, tpi, amu_au
  use io_global, only : stdout, ionode
  use cell_base, only : omega, alat, at, bg, tpiba, tpiba2
  use ions_base, only : nat, atm, zv, tau, ntyp => nsp, ityp
  use gvect,     only : g, ngm, ecutwfc
  use smooth_grid_dimensions, only : nr1s, nr2s, nr3s, nr1sx, nr2sx, nr3sx, nrxxs
  use wvfct,     only : nbnd, npwx, npw, igk, wg, g2kin, current_k  
  use lsda_mod,  only : lsda, isk, nspin
  use klist,     only : xk, nks, nkstot, wk, ngk
  use basis,     only : natomwfc
  use io_files,  only : tmp_dir, prefix, iunat, iunsat, nwordatwfc, find_free_unit
  use mp_global, only : intra_pool_comm, inter_pool_comm
  use mp,        only : mp_sum
  use fft_base,  only : grid_gather, cgather_smooth
  use uspp_param,only : upf 
  use noncollin_module, only : noncolin, npol
  use gsmooth,   only : nls, nlsm
  USE fft_base,       ONLY : dffts
  USE fft_interfaces, ONLY : invfft, fwfft
  

  implicit none
  
  integer,  external  :: atomic_number
  integer             :: ounit , iorb, niorb, index_atom, it
  real(dp)            :: x0(3), ee1(3), ee2(3), ee3(3)
  complex(dp), allocatable  :: psic(:), wfcatom(:,:), psic_BV(:,:,:), phase(:), psic_total(:)
  character (len=256) :: fileout
  character (len=50)  :: mystring, mystring_spin, mystring_ipol
  integer             :: current_spin, ip, istart, iend, ispinor
  real(dp)     :: ktmp, KRcell, H(3,3), HH(3,3)
  real(dp)     :: xkprime(3)
  integer      :: current_cell, i, j, k , m1, m2, icell, jcell, kcell, ik, ispin, icount, ibnd, ncell3_plot(3)
  complex(dp)  :: tmp_phase
  real(dp)     :: sumrho
  integer      :: n, iatom
  real(dp)     :: r_shift(3), atom_position(3, nat*ncell_plot)
  integer      :: atom_type(nat*ncell_plot)
  integer      :: atom_atomicnumber(nat*ncell_plot)
  character(3) :: atom_symbol(nat*ncell_plot)
  integer      :: natom, data_ispin, data_iorbital, data_iatom
  
  call start_clock ('output')
  
  ee1 = 0.d0
  ee2 = 0.d0
  ee3 = 0.d0
  
  if (ionode) then     
     write(stdout, '(A)') '----------------------------------------------------------------------'
     write(stdout, '(A)') '[QO] QO output: start ...'
     write(stdout, '(A)') ''
     ounit  = find_free_unit()
  endif
  
!  nrxxs_total = nr1sx * nr2sx * nr3sx
  allocate( psic_total (nrxxs_total))
  allocate( psic (nrxxs) )
  allocate( psic_BV( nr1sx*ncellx_plot, nr2sx*ncelly_plot, nr3sx*ncellz_plot ) )
  allocate( phase (nrxxs_total) )
  
  ! H for unitcell
  H(:,:) = 0.d0
  H(1:3, 1:3) = transpose(at(1:3,1:3))
  
  ! HH for supercell in Angstrom (plotting)
  ncell3_plot = (/ ncellx_plot, ncelly_plot, ncellz_plot /)
  HH(:,:) = 0.d0
  do i=1,3
     do j=1,3
        HH(j,i) = at(j,i) * dble(ncell3_plot(i)) *alat* bohr_radius_angs
     enddo
  enddo

  !! prepare the atom positions and atom type for netcdf output
  iatom = 1
  natom = ncell_plot * nat
  do i = 1, ncellx_plot
     do j = 1, ncelly_plot
        do k = 1, ncellz_plot
           
           r_shift = at(1:3, 1) * (i-1) * alat*bohr_radius_angs + &
                at(1:3, 2) * (j-1) * alat*bohr_radius_angs + &
                at(1:3, 3) * (k-1) * alat*bohr_radius_angs

           do n=1,nat

              atom_symbol(iatom) = atm(ityp(n))
              atom_type(iatom)   = ityp(n)
              atom_atomicnumber(iatom) = atomic_number(trim(atm(ityp(n))))
              ! positions are in Angstroms
              atom_position(1, iatom) = tau(1,n)*alat*bohr_radius_angs + r_shift(1)
              atom_position(2, iatom) = tau(2,n)*alat*bohr_radius_angs + r_shift(2)
              atom_position(3, iatom) = tau(3,n)*alat*bohr_radius_angs + r_shift(3)
              iatom = iatom + 1
              
           enddo
        end do
     end do
  end do
  
  
  x0     = 0.d0
  ee1(1) = 1.d0
  ee2(2) = 1.d0
  ee3(3) = 1.d0
  !
  ! Output QUASI-ATOMIC ORBITALS (QO)
  !
  ! QO_n(r) = \sum_{k} \sum_{L} u_{nk}(r + L) * exp(i*k*(r+L))
  !         = \sum_{k} u_{nk}(r) * exp(i*k*r) * \sum_{L} * exp(i*k*L)
  !
  do ispin = 1, nspin_general
     
     index_atom = 1
     if (noncolin) then
        niorb = QO_PseudoatomicOrbital_number(index_atom) * 2
     else
        niorb = QO_PseudoatomicOrbital_number(index_atom)
     end if
     
     do iorb = 1, natomwfc
        if (iorb > niorb) then
           index_atom = index_atom + 1
           if (noncolin) then
              niorb = niorb + QO_PseudoatomicOrbital_number(index_atom)*2
           else
              niorb = niorb + QO_PseudoatomicOrbital_number(index_atom)
           end if
        end if
        
        ! Plot out QOs of selected atoms in netcdf format
        if (QO_plot_atomflag(index_atom) > 0) then
           
        psic_BV = cmplx(0.d0, 0.d0)
        
        do ip = 1, npol
           
           istart = npwx*(ip-1)+1
           iend = npwx*ip
           
           ! ----------------------------------------------------------------------------- !
           do ik = 1, nks
              
              current_k = ik
              current_spin = isk(ik)
              if (noncolin) current_spin = 1
              
              if (current_spin == ispin) then
                 
                 ! calculate G vectors at k-point ik
                 call gk_sort (xk (1, ik), ngm, g, ecutwfc / tpiba2, npw, igk, g2kin)
                 g2kin(:) = g2kin(:) * tpiba2
                 
                 npw = ngk (ik)
                 xkprime(1:3) = 0.d0
                 do i = 1, 3
                    do j = 1, 3
                       xkprime(i) = xkprime(i) + xk(j, ik) * at(j, i)
                    end do
                 end do
                 
                 ! phase factor: exp( + i * k * r) :  NOT epr( - ikr) !
                 phase = matmul(Xgrid_total, xk(1:3, ik)) * cmplx(0.d0, 1.d0) * tpi
                 phase = exp(phase)
                 
                 ! read QO wfc (in G-vector representation) on unit iunsat
                 call davcio (QO_wfc, nwordatwfc, iunsat, ik, -1)  
                 
                 ! transform QO contracted array in G space into full array in G space
                 psic_total(:) = 0.d0              
                 psic(:) = cmplx(0.d0, 0.d0)
                 psic(nls(igk(1:npw))) = QO_wfc( istart: (istart+npw-1), iorb)
                 
                 ! transform QO from full array in G space to R space
                 ! call cft3s (psic, nr1s, nr2s, nr3s, nr1sx, nr2sx, nr3sx, 2)
                 call invfft ('Wave', psic, dffts)
#ifdef __PARA
                 call cgather_smooth ( psic, psic_total)
#endif
                 ! write(stdout, *) 'ik, wk =  ', ik, wk(ik)
                 psic_total(1:nrxxs_total) = psic_total(1:nrxxs_total) * wk(ik) * phase(1:nrxxs_total)
                 
                 do icell = - ncellhalfx_plot, ncellhalfx_plot
                    do jcell = - ncellhalfy_plot, ncellhalfy_plot
                       do kcell = - ncellhalfz_plot, ncellhalfz_plot
                          
                          KRcell =  &
                               dble(icell) * xkprime(1)  + &
                               dble(jcell) * xkprime(2)  + &
                               dble(kcell) * xkprime(3)
                          
                          tmp_phase = exp( cmplx(0.d0, 1.d0) * tpi * KRcell)
                          
                          psic_BV( &
                               ((icell+ncellhalfx_plot)*nr1sx+1) : (icell + ncellhalfx_plot +1)*nr1sx, &
                               ((jcell+ncellhalfy_plot)*nr2sx+1) : (jcell + ncellhalfy_plot +1)*nr2sx, &
                               ((kcell+ncellhalfz_plot)*nr3sx+1) : (kcell + ncellhalfz_plot +1)*nr3sx ) &
                               = psic_BV( &
                               ((icell+ncellhalfx_plot)*nr1sx+1) : (icell + ncellhalfx_plot +1)*nr1sx, &
                               ((jcell+ncellhalfy_plot)*nr2sx+1) : (jcell + ncellhalfy_plot +1)*nr2sx, &
                               ((kcell+ncellhalfz_plot)*nr3sx+1) : (kcell + ncellhalfz_plot+1)*nr3sx ) &
                               + tmp_phase * reshape( psic_total, (/nr1sx, nr2sx, nr3sx/) )
                          
                       end do ! do kcell = - ncellhalfz_plot, ncellhalfz_plot
                    end do ! do jcell = - ncellhalfy_plot, ncellhalfy_plot
                 end do !  do icell = - ncellhalfx_plot, ncellhalfx_plot
                 
              end if    ! if (current_spin == ispin) then
           end do    ! do ik = 1, nks
           
#ifdef __PARA
           call mp_sum ( psic_BV, inter_pool_comm )
#endif
           ! here we implicitly assume we are using the symmetric k-point
           ! sampling such that imag part of wavefunctions is cancelled.
           ! psic_BV = dble(psic_BV)
           if (noncolin) then
              psic_BV = psic_BV / sqrt(dble(nr1sx*nr2sx*nr3sx))
           else
              if (nspin == 2) then
                 psic_BV = real(psic_BV) / sqrt(dble(nr1sx*nr2sx*nr3sx))
              else  ! (nspin == 1)
                 psic_BV = real(psic_BV) / sqrt(dble(nr1sx*nr2sx*nr3sx)) / dble(2.d0)
              end if
           end if
           ! ----------------------------------------------------------------------------- !
           if (ionode) then
              
              write(mystring, '(I6)') iorb
              write(mystring_spin, '(I6)') ispin
              write(mystring_ipol, '(I6)') ip
              
              if ( (trim(QO_flag_output_orbital_format) == 'XSF') .OR. &
                   (trim(QO_flag_output_orbital_format) == 'BOTH') ) then
                 
                 !! create XSF file
                 ! xsf file name 
                 fileout = trim(tmp_dir)//trim(prefix)//'.QO_'//trim(adjustl(mystring))//'_spin_'//trim(adjustl(mystring_spin))//'.xsf'
                 if (npol > 1) then
                    fileout = trim(tmp_dir)//trim(prefix)//'.QO_'//trim(adjustl(mystring))//'_spin_'//trim(adjustl(mystring_ipol))//'.xsf'
                 end if
                 write( stdout, '(A)') '  writing QO file to: '//trim(fileout)
                 
                 ! create xsf file
                 open (unit=ounit, file=fileout, form='formatted', status='unknown')
                 
                 ! output xsf structure
                 call xsf_struct_supercell (alat, at, nat, tau, atm, ityp, ounit, (/ncellx_plot, ncelly_plot, ncellz_plot/) )
                 
                 ! output xsf data_3d
                 call xsf_fast_datagrid_3d_supercell &
                      (dble(psic_BV), nr1s*ncellx_plot, nr2s*ncelly_plot, nr3s*ncellz_plot, &
                      nr1sx*ncellx_plot, nr2sx*ncelly_plot, nr3sx*ncellz_plot, &
                      at, alat, ounit, (/ncellx_plot, ncelly_plot, ncellz_plot/) )  
                 
                 ! close xsf file
                 close (unit=ounit)
              end if
                            
              if ( (trim(QO_flag_output_orbital_format) == 'NETCDF') .OR. &
                   (trim(QO_flag_output_orbital_format) == 'BOTH') ) then
                 data_iorbital  = iorb
                 data_ispin = ispin
                 data_iatom = &
                      ncellx_plot*ncelly_plot*ncellhalfz_plot * nat + &
                      ncellx_plot*ncellhalfy_plot * nat + &
                      ncellhalfx_plot * nat + index_atom
                 if (noncolin) data_ispin = 1
                 
                 !! create NetCDF file
                 ! netcdf file name
                 fileout = trim(tmp_dir)//trim(prefix)//'.QO_'//trim(adjustl(mystring))//'_spin_'//trim(adjustl(mystring_spin))//'.nc'
                 if (npol >1) then
                    fileout = trim(tmp_dir)//trim(prefix)//'.QO_'//trim(adjustl(mystring))//'_spinor_'//trim(adjustl(mystring_ipol))//'.nc'
                 end if
                 write( stdout, '(A)') '  writing QO file to: '//trim(fileout)
                 !! write( stdout, *) ' (debug) data_iatom = ', data_iatom
                 call QO_output_netcdf_3d (fileout, psic_BV, &
                      nr1s*ncellx_plot, nr2s*ncelly_plot, nr3s*ncellz_plot, &
                      HH, natom, atom_symbol, atom_type, atom_atomicnumber, atom_position, &
                      data_iorbital, data_ispin, data_iatom, QO_flag_output_orbital_rcutoff )
              end if
              
              sumrho = sum(conjg(psic_BV) * psic_BV)
              write(stdout, '(A, 2i6, 2x, f24.12)') 'ispin, iorb, sumrho : ', ispin, iorb, sumrho
           end if
           
        end do !  do ip = 1, npol
        end if !! if (QO_plot_atomflag(index_atom) > 0)
        
     end do !  do iorb = 1, natomwfc
  end do ! do ispin = 1, nspin
  
  deallocate(psic_BV)
  deallocate(psic)
  deallocate(psic_total)
  deallocate(phase)
  
  if (ionode) then
     write(stdout, '(A)') ''     
     write(stdout, '(A)') '[QO] QO output: end.'
     write(stdout, '(A)') '----------------------------------------------------------------------'
     write(stdout, '(A)') ''     
  end if

  call stop_clock ('output')
  
end subroutine QO_output_orbital



subroutine QO_output_netcdf_3d (nc_file, data, data_dim1, data_dim2, data_dim3, &
     HH, natom, atom_symbol, atom_type, atom_atomicnumber, atom_position, &
     data_iorbital, data_ispin, data_iatom, rcutoff)
!! ================================================================================================== !!
  use netcdf
  use typeSizes
!! ================================================================================================== !!
  
  use kinds,      only : DP, sgl
  use io_files,   only : find_free_unit
  use global_version,       only : version_number
  
  implicit none

  character (len=256), intent (in)    :: nc_file ! NETCDF filename
  integer,  intent(in)     :: data_dim1
  integer,  intent(in)     :: data_dim2
  integer,  intent(in)     :: data_dim3
  complex(dp),  intent(in) :: data(data_dim1, data_dim2, data_dim3)
  real(dp), intent(in)     :: HH(3,3)
  integer,  intent(in)     :: natom
  character(3), intent(in) :: atom_symbol(natom)
  integer,  intent(in)     :: atom_type(natom)
  integer,  intent(in)     :: atom_atomicnumber(natom)
  real(dp), intent(in)     :: atom_position(3, natom)
  integer,  intent(in)     :: data_iorbital
  integer,  intent(in)     :: data_ispin
  integer,  intent(in)     :: data_iatom
  real(dp), intent(in)     :: rcutoff
  
!! ================================================================================================== !!
! NETCDF file_id
  integer  :: nc_file_id 
  
! NETCDF dimensions: length
  integer, parameter :: dim2 = 2, dim3 = 3
  integer, parameter :: real_complex = 2
      
! NETCDF dimensions: ID
  integer  :: dim2_id
  integer  :: dim3_id
  integer  :: real_complex_id
  integer  :: data_dim1_id
  integer  :: data_dim2_id
  integer  :: data_dim3_id
  integer  :: data_fulldim1_id
  integer  :: data_fulldim2_id
  integer  :: data_fulldim3_id
  integer  :: natom_id
  integer  :: data_iatom_id
  integer  :: data_iorbital_id
  integer  :: data_ispin_id
  
! NETCDF variables: ID
  integer  :: supercell_ID
  integer  :: data_region_ID
  integer  :: data_ID
  integer  :: Atom_Symbol_ID
  integer  :: Atom_Type_ID
  integer  :: Atom_AtomicNumber_ID
  integer  :: Atom_Position_ID
!! ================================================================================================== !!
  
  real, allocatable :: data2nc(:,:,:,:)  !! we only need single-precision data for orbital
  real(dp) :: dx, dy, dz
  integer  :: data_region(3, 2), data_fullsize(3)
  integer  :: ix, iy, iz, nx, ny, nz, ii, kk, ll
  
  if (rcutoff <= 0.d0) then
     data_region(1, 1) = 1
     data_region(1, 2) = data_dim1
     data_region(2, 1) = 1
     data_region(2, 2) = data_dim2
     data_region(3, 1) = 1
     data_region(3, 2) = data_dim3
     data_fullsize     = (/ data_dim1, data_dim2, data_dim3 /)
  else
     !! note: currently, reduced NETCDF will NOT work for QO on atoms close to origin
     data_fullsize = (/ data_dim1, data_dim2, data_dim3 /)
     do ii = 1, 3
        dx = HH(ii, ii)/dble(data_fullsize(ii)-1)
        kk = floor  ( (atom_position(ii, data_iatom) - rcutoff) / dx)
        kk = max(kk, 1)
        ll = ceiling( (atom_position(ii, data_iatom) + rcutoff) / dx )
        ll = min(ll, data_fullsize(ii))
        data_region(ii, 1) = kk
        data_region(ii, 2) = ll        
     end do
  end if
  nx = data_region(1, 2) - data_region(1, 1) + 1
  ny = data_region(2, 2) - data_region(2, 1) + 1
  nz = data_region(3, 2) - data_region(3, 1) + 1
  
  allocate( data2nc(nx, ny, nz, real_complex) )
  data2nc(:,:,:,1) =  dble( data( &
       data_region(1, 1):data_region(1, 2), &
       data_region(2, 1):data_region(2, 2), &
       data_region(3, 1):data_region(3, 2) ) )  
  data2nc(:,:,:,2) =  aimag( data( &
       data_region(1, 1):data_region(1, 2), &
       data_region(2, 1):data_region(2, 2), &
       data_region(3, 1):data_region(3, 2) ) )
      
!! CREATE MAIN NETCDF FILE
  nc_file_id = find_free_unit()  
  call check(nf90_create(path = trim(nc_file), cmode = nf90_clobber, ncid = nc_file_id))
  
!! DEFINE THE NETCDF FILE INFORMATION
  call check(nf90_put_att(nc_file_id, nf90_global, "author", &
       "Xiaofeng Qian ( MIT, 2008 )"))
  call check(nf90_put_att(nc_file_id, nf90_global, "title", &
       "NetCDF data file for Quantum-ESPRESSO package  ( version: "//trim(version_number)//" )" ))
  
!! DEFINE DIMENSIONS
  call check( nf90_def_dim(ncid = nc_file_id, name = "dim2",          len=dim2,       dimid=dim2_id      ) )
  call check( nf90_def_dim(ncid = nc_file_id, name = "dim3",          len=dim3,       dimid=dim3_id      ) )
  call check( nf90_def_dim(ncid = nc_file_id, name = "data_dim1",     len=nx,         dimid=data_dim1_id ) )
  call check( nf90_def_dim(ncid = nc_file_id, name = "data_dim2",     len=ny,         dimid=data_dim2_id ) )
  call check( nf90_def_dim(ncid = nc_file_id, name = "data_dim3",     len=nz,         dimid=data_dim3_id ) )
  call check( nf90_def_dim(ncid = nc_file_id, name = "data_fulldim1", len=data_fullsize(1),  dimid=data_fulldim1_id ) )
  call check( nf90_def_dim(ncid = nc_file_id, name = "data_fulldim2", len=data_fullsize(2),  dimid=data_fulldim2_id ) )
  call check( nf90_def_dim(ncid = nc_file_id, name = "data_fulldim3", len=data_fullsize(3),  dimid=data_fulldim3_id ) )
  call check( nf90_def_dim(ncid = nc_file_id, name = "real_complex",  len=real_complex,  dimid=real_complex_id ) )
  call check( nf90_def_dim(ncid = nc_file_id, name = "data_iatom",    len=data_iatom,    dimid=data_iatom_id ) )
  call check( nf90_def_dim(ncid = nc_file_id, name = "data_iorbital", len=data_iorbital, dimid=data_iorbital_id ) )
  call check( nf90_def_dim(ncid = nc_file_id, name = "data_ispin",    len=data_ispin,    dimid=data_ispin_id ) )
  call check( nf90_def_dim(ncid = nc_file_id, name = "natom",         len=natom,         dimid=natom_id ) )
  
!! DEFINE VARIABLES 
!     ATOM
  call check( nf90_def_var( nc_file_id, "Atom_Symbol", nf90_char, &
       dimids= (/ dim3_id, natom_id /), varID= Atom_Symbol_ID) )
  call check( nf90_put_att( nc_file_id,  Atom_Symbol_ID,  "units",  "none") )
  
  call check( nf90_def_var( nc_file_id, "Atom_Type", nf90_int, &
       dimids= (/ natom_id /), varID= Atom_Type_ID) )
  call check( nf90_put_att( nc_file_id,  Atom_Type_ID,  "units",  "none") )
  
  call check( nf90_def_var( nc_file_id, "Atom_AtomicNumber", nf90_int, &
       dimids= (/ natom_id /), varID= Atom_AtomicNumber_ID) ) 
  call check( nf90_put_att( nc_file_id,  Atom_AtomicNumber_ID,  "units",  "none") )  
  
  call check( nf90_def_var( nc_file_id, "Atom_Position", nf90_double, &
       dimids= (/ dim3_id, natom_id /), varID= Atom_Position_ID) )
  call check( nf90_put_att( nc_file_id,  Atom_Position_ID,  "units",  "Angstrom") )
!     UNITCELL
  call check( nf90_def_var( nc_file_id, "supercell", nf90_double, &
       dimids= (/ dim3_id, dim3_id /), varID= supercell_ID) )
  call check( nf90_put_att( nc_file_id,  supercell_ID,  "units",  "Angstrom") )  

  call check( nf90_def_var( nc_file_id, "data_region", nf90_int, &
       dimids= (/ dim3_id, dim2_id /), varID= data_region_ID) )
  call check( nf90_put_att( nc_file_id,  data_region_ID,  "units",  "none") )
  
!     DATA
  call check( nf90_def_var( nc_file_id, "data", nf90_float, &
       dimids= (/ data_dim1_id, data_dim2_id, data_dim3_id, real_complex_id /), varID= data_ID ) )
  call check( nf90_put_att( nc_file_id,  data_ID,  "units",  "none") )
  
!! LEAVE DEFINITION MODE
  call check(nf90_enddef(nc_file_id))
  
!! WRITE VARIABLES
  call check( nf90_put_var( nc_file_id,  Atom_Symbol_ID,   atom_symbol(1:natom) ) )
  call check( nf90_put_var( nc_file_id,  Atom_Type_ID,     atom_type(1:natom) ) )
  call check( nf90_put_var( nc_file_id,  Atom_AtomicNumber_ID,  atom_atomicnumber(1:natom) ) )
  call check( nf90_put_var( nc_file_id,  Atom_Position_ID,      atom_position(1:dim3,1:natom) ) )  
  call check( nf90_put_var( nc_file_id,  supercell_ID,     HH(1:dim3, 1:dim3) ) )
  call check( nf90_put_var( nc_file_id,  data_region_ID,   data_region(1:3, 1:2) ) )
  call check( nf90_put_var( nc_file_id,  data_ID,          data2nc(1:nx, 1:ny, 1:nz, 1:real_complex) ) )
  
!! CLOSE NETCDF FILE  
  call check( nf90_close( nc_file_id ) )
  
  deallocate( data2nc )
  return 
end subroutine QO_output_netcdf_3d

subroutine print_clock_QO
  USE io_global,  ONLY : stdout
  IMPLICIT NONE
  
  write(stdout, '(A)') ''  
  write(stdout, '(A)') '----------------------------------------------------------------------'
  
  call print_clock ('QO')

  write( stdout, * ) '    [ Main routines ]'
  call print_clock ('initialize')
  call print_clock ('construction')
  call print_clock ('tightbinding')
  call print_clock ('write2netcdf')
  call print_clock ('output')
  
  write( stdout, * )
  write( stdout, * ) '    [ General routines ]'
  call print_clock ('init_at_1')
  call print_clock( 'init_us_2' )
  call print_clock ('atomic_wfc')
  call print_clock( 'h_psi' )
  call print_clock( 's_psi' )
  call print_clock( 'calbec' )
  call print_clock( 'cft3s' )
  call print_clock( 'diagh' )
  call print_clock( 'davcio' )
  
  write( stdout, * )
  write( stdout, * ) '    [ Parallel routines ]'
  call print_clock ('fft_scatter')
  write(stdout, '(A)') '----------------------------------------------------------------------'
  write(stdout, '(A)') ''     
  write(stdout, '(A)') ''     
  
end subroutine print_clock_QO

!----------------------------------------------------------------------------
SUBROUTINE QO_openfil()
  !----------------------------------------------------------------------------
  !
  ! ... This routine opens all files needed to the self consistent run,
  ! ... sets various file names, units, record lengths
  !
  USE kinds,          ONLY : DP
  USE wvfct,          ONLY : nbnd, npwx
  USE control_flags,  ONLY:  twfcollect, io_level
  USE io_files,       ONLY : prefix, iunwfc, nwordwfc, iunigk
  USE noncollin_module, ONLY : npol
  USE klist,          ONLY : nks
  USE buffers,        ONLY : get_buffer, save_buffer, open_buffer
  !
  IMPLICIT NONE
  !
  LOGICAL       :: exst, opnd
  !
  twfcollect=.false.
  !
  ! ... nwordwfc is the record length for the direct-access file
  ! ... containing wavefunctions
  !  
  ! nwordwfc = 2 * nbnd * npwx * npol
  nwordwfc = nbnd * npwx * npol
  
  ! ... iunwfc=10: read/write wfc from/to file
  ! ... iunwfc=-1: copy wfc to/from RAM
  IF ( io_level > 0 ) THEN
     iunwfc = 10
  ELSE
     iunwfc = -1
  END IF
  !
  !
  ! CALL diropn( iunwfc, 'wfc', nwordwfc, exst )
  CALL open_buffer( iunwfc, 'wfc', nwordwfc, nks, exst )
  !
  IF ( .NOT. exst ) THEN
     call errore ('QO_openfil','file '//TRIM( prefix )//'.wfc'//' not found',1)     
  END IF
  !
  !
  ! ... iunigk contains the number of PW and the indices igk
  ! ... Note that unit 15 is reserved for error messages
  inquire( UNIT = iunigk, OPENED = opnd )
  CALL seqopn( iunigk, 'igk', 'UNFORMATTED', exst )  
  !
!  IF ( .NOT. exst ) THEN
!     call errore ('QO_openfil','file '//TRIM( prefix )//'.igk'//' not found',1)     
!  END IF
  !
  RETURN
  !
END SUBROUTINE QO_openfil



end module QO_module
