!-----------------------------------------------------------------------------------!
!                                                                                   !
!                            QUASIATOMIC ORBITAL (QO)                               !
!                                                                                   !
!   Description:  Quasiatomic Wannier Functions in Molecules and Solids             !
!                                                                                   !
!   Author:       Xiaofeng Qian  (MIT, Cambridge, MA, USA)                          !
!                                                                                   !
!   Reference:   "Quasiatomic orbitals for ab initio tight-binding analysis",       !
!                 Xiaofeng Qian, Ju Li, Liang Qi, Cai-Zhuang Wang,                  !
!                 Tzu-Liang Chan, Yong-Xin Yao, Kai-Ming Ho, and Sidney Yip,        !
!                 Physical Review B 78, 245112 (2008)                               !
!                                                                                   !
!   Reference (in BIBTEX format):                                                   !
!                                                                                   !
!     @ARTICLE{QianLQWCYHY08,                                                       !
!          author    = {Qian, Xiaofeng and Li, Ju and Qi, Liang and Wang, C. Z. and !
!                       Chan, T. L. and Yao, Y. X. and Ho, K. M. and Yip, Sidney},  !
!          title     = {Quasiatomic orbitals for ab initio tight-binding analysis}, ! 
!          journal   = {Phys. Rev. B},                                              !
!          year	     = {2008},                                                      !
!          volume    = {78},                                                        !
!          pages     = {245112}                                                     !
!     }                                                                             !
!                                                                                   !
!   Please cite the above reference if you use the code or part of the code.        !  
!                                                                                   !
!   Copyright (C) 2009 Xiaofeng Qian                                                !
!                                                                                   !
!   This file is distributed under the terms of the                                 !
!   GNU General Public License. See the file `License'                              !
!   in the root directory of the present distribution,                              !
!   or http://www.gnu.org/copyleft/gpl.txt .                                        !
!                                                                                   !
!-----------------------------------------------------------------------------------!
!
     SUBROUTINE QO_readcards
       !----------------------------------------------------------------------
       !
       USE kinds,     ONLY : DP
       USE io_global, ONLY : stdout 
       USE constants, ONLY : angstrom_au
       USE parser,    ONLY : field_count, read_line
       USE io_global, ONLY : ionode, ionode_id
       !
       USE input_parameters
       !
       IMPLICIT NONE
       !
       CHARACTER(LEN=2)           :: prog   ! calling program ( FP, PW, CP )
       CHARACTER(LEN=256)         :: input_line
       CHARACTER(LEN=80)          :: card
       CHARACTER(LEN=1), EXTERNAL :: capital
       LOGICAL                    :: tend, end_of_file
       INTEGER                    :: i
       
       
100    CALL read_line( input_line, end_of_file=tend )
       !
       IF( tend ) GO TO 120
       IF( input_line == ' ' .OR. input_line(1:1) == '#' ) GO TO 100
       !
       READ (input_line, *) card
       !
       DO i = 1, LEN_TRIM( input_line )
          input_line( i : i ) = capital( input_line( i : i ) )
       END DO
       
       if ( trim(card) == 'AO_CHANNEL_REMOVE' ) then
          call QO_card_atomicorbital_channel_remove( input_line )
       else if ( trim(card) == 'AO_CHANNEL_ADD' ) then
          call QO_card_atomicorbital_channel_add( input_line )
       else if (trim(card) == 'AO_RESCALING') then
          call QO_card_atomicorbital_rescaling( input_line )
       else if (trim(card) == 'QO_PLOT_CELL_DIMENSION') then
          call QO_card_plot_cell_dimension( input_line )
       else if (trim(card) == 'QO_PLOT_ATOMS') then
          call QO_card_plot_atoms( input_line )
!       else if (trim(card) == 'QO_PLOT_KPOINTS') then
!          call QO_card_plot_kpoints( input_line )
       else if (trim(card) == 'QO_PLOT_KPOINTS_INDEX') then
          call QO_card_plot_kpoints_index( input_line )
       else
       end if
       
       GOTO 100
       
120    CONTINUE
       
       call QO_bcast_card()
       
       RETURN
       
     end SUBROUTINE QO_readcards
     

     !------------------------------------------------------------------------
     !    BEGIN manual
     !----------------------------------------------------------------------
     !
     !  AO_RESCALING
     !
     !   rescaling the original pseudo-atomic orbitals from pseudopotential library
     !
     ! Syntax:
     !
     !    AO_RESCALING
     !      label(1)    rescaling(1) ... rescaling(6)
     !       ...        ...        ...
     !      label(n)    rescaling(1) ... rescaling(6)
     !
     ! Example:
     !
     ! AO_RESCALING
     !  O 0.1 0.1  0. 0. 0. 0.
     !  H 0.  0.   0. 0. 0. 0.
     !
     ! Where:
     !
     !      label(i)  ( character(len=4) )  label of the atomic species 
     !      rescaling(i)   ( real )         radial_ao(i) * exp( - rescaling(i) * r )
     !
     !----------------------------------------------------------------------
     !    END manual
     !------------------------------------------------------------------------
     !
     SUBROUTINE QO_card_atomicorbital_rescaling( input_line )
       !
       use wrappers,   only : feval_infix
       use parser,     only : field_count, read_line, get_field
       use ions_base,  only : ntyp => nsp
       use input_parameters, only : atom_label, nsx, taspc
       use QO_module
       
       IMPLICIT NONE
       !
       CHARACTER(LEN=256) :: input_line
       INTEGER            :: is, ip, k, ifield, nfield, ierr
       CHARACTER(LEN=4)   :: lb_pos
       CHARACTER(LEN=256) :: psfile
       LOGICAL, SAVE      :: tread = .FALSE.
       CHARACTER(len=256) :: field_str, error_msg
       !
       !
       IF ( tread ) THEN
          CALL errore( ' card_atomicorbital_rescaling  ', ' two occurrence ', 2 )
       END IF
       IF ( ntyp > nsx ) THEN
         CALL errore( ' card_atomicorbital_rescaling ', ' nsp out of range ', ntyp )
       END IF
       
       ! write (*, *) 'ntyp = ', ntyp
       
       DO is = 1, ntyp
          !
          CALL read_line( input_line )
          
          CALL field_count( nfield, input_line )
          
          ! write(*, *) input_line
          ! READ( input_line, * ) &
          !     lb_pos, &
          !     ( QO_PseudoatomicOrbital_rescalingfactor(is, k), k = 1, QO_PseudoatomicOrbital_nlm_effective(is) )
          
          ! atom_pfile(is) = TRIM( psfile )
          ! read atom symbol (column 1) and coordinate
          CALL get_field(1, lb_pos, input_line)
          lb_pos         = ADJUSTL( lb_pos )
          atom_label(is) = TRIM( lb_pos )
          
          DO ip = 1, is - 1
             IF ( atom_label(ip) == atom_label(is) ) THEN
                CALL errore( ' card_atomicorbital_rescaling ', &
                           & ' two occurrences of the same atomic label ', is )
             END IF
          END DO
          
          error_msg = 'Error while parsing AO_RESCALING card.'
          do ifield = 1, nfield-1
             CALL get_field(ifield+1, field_str, input_line)
             QO_PseudoatomicOrbital_rescalingfactor(is, ifield) = feval_infix(ierr, field_str )
             CALL errore('card_AO_RESCALING', error_msg, ierr)
          end do
!!$          write (*, *) 'read', is, atom_label(is), QO_PseudoatomicOrbital_rescalingfactor(is, 1:nchix), QO_PseudoatomicOrbital_nlm_effective(is)
          
       END DO
       taspc = .TRUE.
       tread = .TRUE.
       !
       RETURN
       !
     END SUBROUTINE QO_card_atomicorbital_rescaling

     !------------------------------------------------------------------------
     !    BEGIN manual
     !----------------------------------------------------------------------
     !
     !  AO_CHANNEL
     !
     !   rescaling the original pseudo-atomic orbitals from pseudopotential library
     !
     ! Syntax:
     !
     !    AO_CHANNEL
     !      label(1)    channel_flag(1) ... channel_flag(6)
     !       ...        ...        ...
     !      label(n)    channel_flag(1) ... channel_flag(6)
     !
     ! Example:
     !
     ! AO_CHANNEL
     !  O 1  1  0  0  0  0
     !  H 1  0  0  0  0  0
     !
     ! Where:
     !
     !      label(i)  ( character(len=4) )  label of the atomic species 
     !      channel_flag(i)    ( integer )      if >0,  then add channel (l-quantum number)
     !                                          if ==0, then turn on the channel
     !                                          if <0,  then turn off the EXISTING channel
     !                                             
     !
     !----------------------------------------------------------------------
     !    END manual
     !------------------------------------------------------------------------
     !
     SUBROUTINE QO_card_atomicorbital_channel_add( input_line )
       !
       use wrappers,   only : feval_infix
       use parser,     only : field_count, read_line, get_field
       use ions_base,  only : ntyp => nsp
       use input_parameters, only : atom_label, nsx, taspc
       use QO_module
       
       IMPLICIT NONE
       !
       CHARACTER(LEN=256) :: input_line
       INTEGER            :: is, ip, k, ifield, nfield, ierr
       CHARACTER(LEN=4)   :: lb_pos
       CHARACTER(LEN=256) :: psfile
       LOGICAL, SAVE      :: tread = .FALSE.
       CHARACTER(len=256) :: field_str, error_msg
       !
       !
       IF ( tread ) THEN
          CALL errore( ' card_atomicorbital_channel_add  ', ' two occurrence ', 2 )
       END IF
       IF ( ntyp > nsx ) THEN
         CALL errore( ' card_atomicorbital_channel_add ', ' nsp out of range ', ntyp )
       END IF
       
       ! write (*, *) 'ntyp = ', ntyp
       
       DO is = 1, ntyp
          !
          CALL read_line( input_line )
          
          CALL field_count( nfield, input_line )
          
          ! write(*, *) input_line
          ! READ( input_line, * ) &
          !     lb_pos, &
          !     ( QO_PseudoatomicOrbital_rescalingfactor(is, k), k = 1, QO_PseudoatomicOrbital_nlm_effective(is) )
          
          ! atom_pfile(is) = TRIM( psfile )
          ! read atom symbol (column 1) and coordinate
          CALL get_field(1, lb_pos, input_line)
          lb_pos         = ADJUSTL( lb_pos )
          atom_label(is) = TRIM( lb_pos )
          
          DO ip = 1, is - 1
             IF ( atom_label(ip) == atom_label(is) ) THEN
                CALL errore( ' card_atomicorbital_channel_add ', &
                           & ' two occurrences of the same atomic label ', is )
             END IF
          END DO
          
          error_msg = 'Error while parsing AO_CHANNEL_ADD card.'
          do ifield = 1, nfield-1
             CALL get_field(ifield+1, field_str, input_line)
             QO_PseudoatomicOrbital_channel_add(is, ifield) = feval_infix(ierr, field_str )
             CALL errore('card_AO_CHANNEL_ADD', error_msg, ierr)
          end do
!!$          write (*, *) 'read', is, atom_label(is), QO_PseudoatomicOrbital_rescalingfactor(is, 1:nchix), QO_PseudoatomicOrbital_nlm_effective(is)
          
       END DO
       taspc = .TRUE.
       tread = .TRUE.
       !
       RETURN
       !
     END SUBROUTINE QO_card_atomicorbital_channel_add

     
     
     
     SUBROUTINE QO_card_atomicorbital_channel_remove( input_line )
       !
       use wrappers,   only : feval_infix
       use parser,     only : field_count, read_line, get_field
       use ions_base,  only : ntyp => nsp
       use input_parameters, only : atom_label, nsx, taspc
       use QO_module
       
       IMPLICIT NONE
       !
       CHARACTER(LEN=256) :: input_line
       INTEGER            :: is, ip, k, ifield, nfield, ierr
       CHARACTER(LEN=4)   :: lb_pos
       CHARACTER(LEN=256) :: psfile
       LOGICAL, SAVE      :: tread = .FALSE.
       CHARACTER(len=256) :: field_str, error_msg
       !
       !
       IF ( tread ) THEN
          CALL errore( ' card_atomicorbital_channel_remove ', ' two occurrence ', 2 )
       END IF
       IF ( ntyp > nsx ) THEN
         CALL errore( ' card_atomicorbital_channel_remove ', ' nsp out of range ', ntyp )
       END IF
       
       DO is = 1, ntyp
          !
          CALL read_line( input_line )
          
          CALL field_count( nfield, input_line )
          
          ! atom_pfile(is) = TRIM( psfile )
          ! read atom symbol (column 1) and coordinate
          CALL get_field(1, lb_pos, input_line)
          lb_pos         = ADJUSTL( lb_pos )
          atom_label(is) = TRIM( lb_pos )
          
          DO ip = 1, is - 1
             IF ( atom_label(ip) == atom_label(is) ) THEN
                CALL errore( ' card_atomicorbital_channel ', &
                           & ' two occurrences of the same atomic label ', is )
             END IF
          END DO
          
          error_msg = 'Error while parsing AO_CHANNEL_REMOVE card.'
          do ifield = 1, nfield-1
             CALL get_field(ifield+1, field_str, input_line)
             QO_PseudoatomicOrbital_channel_remove(is, ifield) = feval_infix(ierr, field_str )
             CALL errore('card_AO_CHANNEL_REMOVE', error_msg, ierr)
          end do
!!$          write (*, *) 'read', is, atom_label(is), QO_PseudoatomicOrbital_rescalingfactor(is, 1:nchix), QO_PseudoatomicOrbital_nlm_effective(is)
          
       END DO
       taspc = .TRUE.
       tread = .TRUE.
       !
       RETURN
       !
     END SUBROUTINE QO_card_atomicorbital_channel_remove

     
     
     
     SUBROUTINE QO_card_plot_cell_dimension( input_line )
       !
       USE parser,    ONLY : field_count, read_line
       use ions_base,  only : ntyp => nsp
       USE input_parameters, only : atom_label, nsx, taspc
       use QO_module
       
       IMPLICIT NONE
       !
       CHARACTER(LEN=256) :: input_line
       INTEGER            :: is, ip, k
       CHARACTER(LEN=256) :: psfile
       LOGICAL, SAVE      :: tread = .FALSE.
       !
       !
       IF ( tread ) THEN
          CALL errore( ' card_plot_cell_dimension ', ' two occurrence ', 2 )
       END IF
       
       CALL read_line( input_line )
       ! write(*, *) input_line
       READ( input_line, * )  ( QO_plot_cell_dimension(k), k = 1, 3 )
       ncellx_plot = QO_plot_cell_dimension(1)
       ncelly_plot = QO_plot_cell_dimension(2)
       ncellz_plot = QO_plot_cell_dimension(3)
       ncellhalfx_plot = floor(dble(ncellx_plot-1)/2.d0)
       ncellhalfy_plot = floor(dble(ncelly_plot-1)/2.d0)
       ncellhalfz_plot = floor(dble(ncellz_plot-1)/2.d0)
       ncellx_plot = 2*ncellhalfx_plot + 1
       ncelly_plot = 2*ncellhalfy_plot + 1
       ncellz_plot = 2*ncellhalfz_plot + 1
       ncell_plot  = ncellx_plot * ncelly_plot * ncellz_plot       
       
       taspc = .TRUE.
       tread = .TRUE.
       !
       RETURN
       !
     END SUBROUTINE QO_card_plot_cell_dimension
     
     
     SUBROUTINE QO_card_plot_atoms( input_line )
       !
       USE parser,    ONLY : field_count, read_line
       use ions_base,  only : ntyp => nsp
       USE input_parameters, only : atom_label, nsx, taspc
       use QO_module
       
       IMPLICIT NONE
       !
       CHARACTER(LEN=256) :: input_line
       INTEGER            :: is, ip, k, na, natom
       CHARACTER(LEN=256) :: psfile
       LOGICAL, SAVE      :: tread = .FALSE.
       !
       !
       IF ( tread ) THEN
          CALL errore( ' card_plot_atoms ', ' two occurrence ', 2 )
       END IF
       
       CALL read_line( input_line )
       ! write(*, *) input_line
       READ( input_line, * )  na
       natom = QO_plot_natom
       if ((na <= QO_plot_natom) .and. (na > 0)) then
          do k = 1, QO_plot_natom
             QO_plot_atomflag( k ) = 0
          end do
          QO_plot_natom = na
          if (allocated(QO_plot_atomindex)) then
             deallocate(QO_plot_atomindex)
          end if
          allocate( QO_plot_atomindex(QO_plot_natom  ) )
          CALL read_line( input_line )
!          write(*, *) input_line
          READ( input_line, * )  (QO_plot_atomindex(k), k = 1, QO_plot_natom)
          do k = 1, QO_plot_natom
             if (QO_plot_atomindex(k) <= natom) then
                QO_plot_atomflag( QO_plot_atomindex(k) ) = 1
             end if
          end do
       end if
       
       taspc = .TRUE.
       tread = .TRUE.
       !
       RETURN
       !
     END SUBROUTINE QO_card_plot_atoms
     
     SUBROUTINE QO_card_plot_kpoints( input_line )
       !
       USE parser,    ONLY : field_count, read_line
       use ions_base,  only : ntyp => nsp
       USE input_parameters, only : atom_label, nsx, taspc
       use QO_module
       
       IMPLICIT NONE
       !
       CHARACTER(LEN=256) :: input_line
       INTEGER            :: is, ip, ik, nkpoint
       CHARACTER(LEN=256) :: psfile
       LOGICAL, SAVE      :: tread = .FALSE.
       !
       !
       IF ( tread ) THEN
          CALL errore( ' card_plot_cell_dimension ', ' two occurrence ', 2 )
       END IF
       
       CALL read_line( input_line )
       ! write(*, *) input_line
       READ( input_line, * )  nkpoint
       if (allocated(QO_plot_kpoints)) then
          deallocate(QO_plot_kpoints)
       end if
       
       QO_plot_number_of_kpoints = 0
       if (nkpoint >=1) then
          QO_plot_number_of_kpoints = nkpoint
          allocate( QO_plot_kpoints(nkpoint,3) )
       else
          return
       end if
       
       DO ik = 1, nkpoint
          CALL read_line( input_line )
          READ( input_line, * )  (QO_plot_kpoints(ik, ip), ip = 1, 3)
       END DO
       
       taspc = .TRUE.
       tread = .TRUE.
       !
       RETURN
       !
     END SUBROUTINE QO_card_plot_kpoints
     
     
     SUBROUTINE QO_card_plot_kpoints_index( input_line )
       !
       USE parser,    ONLY : field_count, read_line
       use ions_base,  only : ntyp => nsp
       USE input_parameters, only : atom_label, nsx, taspc
       use QO_module
       
       IMPLICIT NONE
       !
       CHARACTER(LEN=256) :: input_line
       INTEGER            :: is, ip, ik, nkpoint
       CHARACTER(LEN=256) :: psfile
       LOGICAL, SAVE      :: tread = .FALSE.
       !
       !
       IF ( tread ) THEN
          CALL errore( ' card_plot_cell_dimension ', ' two occurrence ', 2 )
       END IF
       
       CALL read_line( input_line )
       ! write(*, *) input_line
       READ( input_line, * )  nkpoint
       if (allocated(QO_plot_kpoints_index)) then
          deallocate(QO_plot_kpoints_index)
       end if
       
       QO_plot_number_of_kpoints = 0
       if (nkpoint >=1) then
          QO_plot_number_of_kpoints = nkpoint
          allocate( QO_plot_kpoints_index(nkpoint) )
       else
          return
       end if
       
       CALL read_line( input_line )
       READ( input_line, * )  (QO_plot_kpoints_index(ik), ik = 1, nkpoint)
       
       taspc = .TRUE.
       tread = .TRUE.
       !
       RETURN
       !
     END SUBROUTINE QO_card_plot_kpoints_index
     
     
     subroutine QO_bcast_card
    !-----------------------------------------------------------------------
    ! Broadcast input data to all processors 
    !-----------------------------------------------------------------------
#ifdef __PARA
#include "f_defs.h"
      USE io_files,      ONLY : prefix, tmp_dir
      USE us,            ONLY : spline_ps
      USE mp,            ONLY : mp_bcast, mp_end, mp_bcast_rm, mp_sum
      USE mp_global,     ONLY : inter_pool_comm, intra_pool_comm 
      USE QO_module
      
      implicit none
      integer :: root = 0
      
      CALL mp_bcast( QO_plot_atomindex, root )
      CALL mp_bcast( QO_plot_atomflag, root )
      CALL mp_bcast( QO_plot_natom, root)
      call mp_bcast( ncellx_plot, root)
      call mp_bcast( ncelly_plot, root)
      call mp_bcast( ncellz_plot, root)
      call mp_bcast( ncellhalfx_plot, root)
      call mp_bcast( ncellhalfy_plot, root)
      call mp_bcast( ncellhalfz_plot, root)
      call mp_bcast( ncellx_plot, root)
      call mp_bcast( ncelly_plot, root)
      call mp_bcast( ncellz_plot, root)
      call mp_bcast( ncell_plot, root)
      call mp_bcast( QO_PseudoatomicOrbital_channel_add, root)
      call mp_bcast( QO_PseudoatomicOrbital_channel_remove, root)
      call mp_bcast( QO_PseudoatomicOrbital_rescalingfactor, root)
      call mp_bcast( QO_plot_cell_dimension, root  )
      call mp_bcast( QO_plot_number_of_kpoints, root )
      if (QO_plot_number_of_kpoints > 0) then
         call mp_bcast( QO_plot_kpoints_index, root)
      end if
#endif
    end subroutine QO_bcast_card
